#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import copy
import warnings
from typing import List, Optional, Union

import joblib
import numpy as np
import numpy.typing as npt
import scipy.linalg
import torch

import cebra.data.base as cebra_data_base
import cebra.data.multi_session as cebra_data_multisession
import cebra.data.single_session as cebra_data_singlesession


def get_loader_options(dataset: cebra_data_base.Dataset) -> List[str]:
    """Return all possible dataloaders for the given dataset."""

    loader_options = []
    if isinstance(dataset, cebra_data_singlesession.SingleSessionDataset):
        mixed = True
        if dataset.continuous_index is not None:
            loader_options.append(cebra_data_singlesession.ContinuousDataLoader)
        else:
            mixed = False
        if dataset.discrete_index is not None:
            loader_options.append(cebra_data_singlesession.DiscreteDataLoader)
        else:
            mixed = False
        if mixed:
            loader_options.append(cebra_data_singlesession.MixedDataLoader)
    elif isinstance(dataset, cebra_data_multisession.MultiSessionDataset):
        mixed = True
        if dataset.continuous_index is not None:
            loader_options.append(
                cebra_data_multisession.ContinuousMultiSessionDataLoader)
        else:
            mixed = False
        if dataset.discrete_index is not None:
            pass  # not implemented yet
        else:
            mixed = False
        if mixed:
            pass  # not implemented yet
    else:
        raise TypeError(f"Invalid dataset type: {dataset}")
    return loader_options


def _require_numpy_array(array: Union[npt.NDArray, torch.Tensor]):
    if not isinstance(array, np.ndarray):
        if isinstance(array, torch.Tensor):
            array = array.numpy()
        else:
            raise ValueError(
                f"Invalid dtype for the provided embedding, got a {type(array)}."
            )
    elif len(array.shape) == 1:
        array = np.expand_dims(array, axis=1)

    return array


class OrthogonalProcrustesAlignment:
    """Aligns two dataset by solving the orthogonal Procrustes problem.

    Tip:
        In linear algebra, the orthogonal Procrustes problem is a matrix approximation
        problem. Considering two matrices A and B, it consists in finding the orthogonal
        matrix R which most closely maps A to B, so that it minimizes the Frobenius norm of
        ``(A @ R) - B`` subject to ``R.T @ R = I``.
        See :py:func:`scipy.linalg.orthogonal_procrustes` for more information.

    For each dataset, the data and labels to align the data on is provided.

    1. The ``top_k`` indexes of the labels to align (``label``) that are the closest to
       the labels of the reference dataset (``ref_label``) are selected and used to sample
       from the dataset to align (``data``).
    2. ``data`` and ``ref_data`` (the reference dataset) are subsampled to the same number
       of samples ``subsample``.
    3. The orthogonal mapping is computed, using :py:func:`scipy.linalg.orthogonal_procrustes`,
       on those subsampled datasets.
    4. The resulting orthongonal matrix ``_transform`` can be used to map the original ``data``
       to the ``ref_data``.

    Note:
        ``data`` and ``ref_data`` can be of different sample size (axis 0) but **must** have the same number
        of features (axis 1) to be aligned.

    Attributes:
        top_k (int): Number of indexes in the labels of the matrix to align to consider for alignment
            (``label``). The selected indexes consist in the ``top_k`` th indexes the closest to the
            reference labels (``ref_label``).
        subsample (int): Number of samples to subsample the ``data`` and ``ref_data`` from, to solve the orthogonal
            Procrustes problem on.
    """

    def __init__(self, top_k: int = 5, subsample: Optional[int] = None):
        self.subsample = subsample
        self.top_k = top_k

    def _distance(self, label_i: npt.NDArray,
                  label_j: npt.NDArray) -> npt.NDArray:
        """Compute the Euclidean distance between two matrices.

        Args:
            label_i: An array.
            label_j: A second array.

        Returns:
            The Euclidean distance between ``label_i`` and ``label_j``.
        """
        norm_i = (label_i**2).sum(1)
        norm_j = (label_j**2).sum(1)
        diff = np.einsum("nd,md->nm", label_i, label_j)
        diff = norm_i[:, None] + norm_j[None, :] - 2 * diff
        return diff

    def fit(
        self,
        ref_data: Union[npt.NDArray, torch.Tensor],
        data: Union[npt.NDArray, torch.Tensor],
        ref_label: Optional[npt.NDArray] = None,
        label: Optional[npt.NDArray] = None,
    ) -> "OrthogonalProcrustesAlignment":
        """Compute the matrix solution of the orthogonal Procrustes problem.

        The obtained matrix is used to align a dataset to a reference dataset.

        Args:
            ref_data: Reference data matrix on which to align the data.
            data: Data matrix to align on the reference dataset.
            ref_label: Set of indices associated to ``ref_data``.
            label: Set of indices associated to ``data``.

        Returns:
            ``self``, for chaining operations.

        Example:

            >>> import cebra.data.helper
            >>> import numpy as np
            >>> ref_embedding = np.random.uniform(0, 1, (1000, 30))
            >>> aux_embedding = np.random.uniform(0, 1, (800, 30))
            >>> ref_label = np.random.uniform(0, 1, (1000, 1))
            >>> aux_label = np.random.uniform(0, 1, (800, 1))
            >>> orthogonal_procrustes = cebra.data.helper.OrthogonalProcrustesAlignment()
            >>> orthogonal_procrustes = orthogonal_procrustes.fit(ref_data=ref_embedding,
            ...                                                   data=aux_embedding,
            ...                                                   ref_label=ref_label,
            ...                                                   label=aux_label)

        """
        if len(ref_data.shape) == 1:
            ref_data = np.expand_dims(ref_data, axis=1)
        if len(data.shape) == 1:
            data = np.expand_dims(data, axis=1)

        if ref_data.shape[0] == data.shape[
                0] and ref_label is None and label is None:
            # data are already aligned across timepoints, so we can use them as labels
            n_samples = ref_data.shape[0]
            ref_label = np.expand_dims(np.arange(n_samples), axis=1)
            label = np.expand_dims(np.arange(n_samples), axis=1)
        elif ref_data.shape[0] == data.shape[0] and (ref_label is None or
                                                     label is None):
            raise ValueError(
                "Missing labels: the data to align are the same shape but you provided only "
                "one of the sets of labels. Either provide both the reference and alignment "
                "labels or none.")
        else:
            if ref_label is None or label is None:
                raise ValueError(
                    "Missing labels: the data to align are not the same shape, "
                    "provide labels to align the data and reference data.")

            if len(ref_label.shape) == 1:
                ref_label = np.expand_dims(ref_label, axis=1)
            if len(label.shape) == 1:
                label = np.expand_dims(label, axis=1)

        if ref_data.shape[0] < self.top_k:
            raise ValueError(
                f"Invalid reference data, reference data should at least "
                f"contain {self.top_k} samples, got {ref_data.shape[0]}.")

        if ref_data.shape[0] != ref_label.shape[0]:
            raise ValueError(
                f"Mismatched reference data and associated reference labels: "
                f"they should contain the same number of samples, but "
                f"got ref_data:{ref_data.shape[0]} samples and ref_labels:{ref_label.shape[0]} samples."
            )

        if data.shape[0] != label.shape[0]:
            raise ValueError(
                f"Mismatched reference data and associated reference labels: "
                f"they should contain the same number of samples, but "
                f"got ref_data:{data.shape[0]} samples and ref_labels:{label.shape[0]} samples."
            )

        distance = self._distance(label, ref_label)

        # keep indexes of the {self.top_k} labels the closest to the reference labels
        target_idx = np.argsort(distance, axis=1)[:, :self.top_k]

        if data.shape[1] != ref_data.shape[1]:
            raise ValueError(
                "Invalid data: reference data and data should have the same number of features but "
                f"got data: {data.shape[1]} features and ref_data: {ref_data.shape[1]} features."
            )

        # Get the whole data to align and only the selected closest samples
        # from the reference data.
        X = data[:, None].repeat(self.top_k, axis=1).reshape(-1, data.shape[1])
        Y = ref_data[target_idx].reshape(-1, ref_data.shape[1])

        # Augment data and reference data so that same size
        if self.subsample is not None:
            if self.subsample > len(X):
                warnings.warn(
                    f"The number of datapoints in the dataset ({len(X)}) "
                    f"should be larger than the 'subsample' "
                    f"parameter ({self.subsample}). Ignoring subsampling and "
                    f"computing alignment on the full dataset instead, which will "
                    f"give better results.")
            else:
                if self.subsample < 1000:
                    warnings.warn(
                        "This function is experimental when the subsample dimension "
                        "is less than 1000. You can probably use the whole dataset "
                        "for alignment by setting subsample=None.")

                idc = np.random.choice(len(X), self.subsample)
                X = X[idc]
                Y = Y[idc]

        # Compute orthogonal matrix that most closely maps X to Y using the orthogonal Procrustes problem.
        self._transform, _ = scipy.linalg.orthogonal_procrustes(X, Y)

        return self

    def transform(self, data: Union[npt.NDArray, torch.Tensor]) -> npt.NDArray:
        """Transform the data using the matrix solution computed in py:meth:`fit`.

        Args:
            data: The 2D data matrix to align.

        Returns:
            The aligned input matrix.

        Example:

            >>> import cebra.data.helper
            >>> import numpy as np
            >>> ref_embedding = np.random.uniform(0, 1, (1000, 30))
            >>> aux_embedding = np.random.uniform(0, 1, (800, 30))
            >>> ref_label = np.random.uniform(0, 1, (1000, 1))
            >>> aux_label = np.random.uniform(0, 1, (800, 1))
            >>> orthogonal_procrustes = cebra.data.helper.OrthogonalProcrustesAlignment()
            >>> orthogonal_procrustes = orthogonal_procrustes.fit(ref_data=ref_embedding,
            ...                                                   data=aux_embedding,
            ...                                                   ref_label=ref_label,
            ...                                                   label=aux_label)
            >>> aligned_aux_embedding = orthogonal_procrustes.transform(data=aux_embedding)
            >>> assert aligned_aux_embedding.shape == aux_embedding.shape

        """
        if self.transform is None:
            raise ValueError("Call fit() first.")

        if len(data.shape) == 1:
            data = np.expand_dims(data, axis=1)

        return data @ self._transform

    def fit_transform(
        self,
        ref_data: npt.NDArray,
        data: npt.NDArray,
        ref_label: Optional[npt.NDArray] = None,
        label: Optional[npt.NDArray] = None,
    ) -> npt.NDArray:
        """Compute the matrix solution to align a data array to a reference matrix.

        Note:
            Uses a combination of :py:meth:`~.OrthogonalProcrustesAlignment.fit` and :py:meth:`~.OrthogonalProcrustesAlignment.transform`.


        Args:
            ref_data: Reference data matrix on which to align the data.
            data: Data matrix to align on the reference dataset.
            ref_label: Set of indices associated to ``ref_data``.
            label: Set of indices associated to ``data``.

        Returns:
            The ``data`` matrix aligned onto the reference data matrix.

        Example:

            >>> import cebra.data.helper
            >>> import numpy as np
            >>> ref_embedding = np.random.uniform(0, 1, (1000, 30))
            >>> aux_embedding = np.random.uniform(0, 1, (800, 30))
            >>> ref_label = np.random.uniform(0, 1, (1000, 1))
            >>> aux_label = np.random.uniform(0, 1, (800, 1))
            >>> orthogonal_procrustes = cebra.data.helper.OrthogonalProcrustesAlignment(top_k=10,
            ...                                                                         subsample=700)
            >>> aligned_aux_embedding = orthogonal_procrustes.fit_transform(ref_data=ref_embedding,
            ...                                                             data=aux_embedding,
            ...                                                             ref_label=ref_label,
            ...                                                             label=aux_label)
            >>> assert aligned_aux_embedding.shape == aux_embedding.shape

        """
        self.fit(ref_data, data, ref_label, label)
        return self.transform(data)


def ensemble_embeddings(
    embeddings: List[Union[npt.NDArray, torch.Tensor]],
    labels: Optional[List[Union[npt.NDArray, torch.Tensor]]] = None,
    post_norm: bool = False,
    n_jobs: int = 0,
) -> npt.NDArray:
    """Ensemble aligned embeddings together.

    The embeddings contained in ``embeddings`` are aligned onto the same embedding,
    using :py:class:`OrthogonalProcrustesAlignment`. Then, they are averaged and the
    resulting averaged embedding is the ensemble embedding.

    Tip:
        By ensembling embeddings coming from the same dataset but obtained from different models,
        the resulting joint embedding usually shows an increase in performances compared to the individual
        embeddings.

    Note:
        The embeddings in ``embeddings`` must be the same shape, i.e., the same number of samples
        and same number of features (axis 1).

    Args:
        embeddings: List of embeddings to align and ensemble.
        labels: Optional list of indexes associated to the embeddings in ``embeddings`` to align the embeddings on.
            To be ensembled, the embeddings should already be aligned on time, and consequently do not require extra
            labels for alignment.
        post_norm: If True, the resulting joint embedding is normalized (divided by its norm across
            the features - axis 1).
        n_jobs: The maximum number of concurrently running jobs to compute embedding alignment in a parallel manner using
            :py:class:`joblib.Parallel`. Specify ``0`` to iterate naively over the embeddings for ensembling without using
            :py:class:`joblib.Parallel`. Specify ``-1`` to use all cores. Using more than a single core can considerably
            speed up the computation of ensembled embeddings for large datasets, but will also require more memory.

    Returns:
        A :py:func:`numpy.array` corresponding to the joint embedding.

    Example:

        >>> import cebra.data.helper
        >>> import numpy as np
        >>> embedding1 = np.random.uniform(0, 1, (100, 4))
        >>> embedding2 = np.random.uniform(0, 1, (100, 4))
        >>> embedding3 = np.random.uniform(0, 1, (100, 4))
        >>> joint_embedding = cebra.data.helper.ensemble_embeddings(embeddings=[embedding1, embedding2, embedding3])
        >>> assert joint_embedding.shape == embedding1.shape

    """
    if n_jobs is None or not isinstance(n_jobs, int):
        raise ValueError(
            "Invalid n_jobs: it should be an Integer, got {n_jobs}. To not use parallelisation, set n_jobs to 0."
        )

    embeddings = [_require_numpy_array(embedding) for embedding in embeddings]

    # align each embedding from the list to the first one
    ortho_alignment = OrthogonalProcrustesAlignment()
    for embedding in embeddings[1:]:
        if embeddings[0].shape != embedding.shape:
            raise ValueError(
                f"Inconsistent embeddings shapes, one embedding is {embedding.shape}, another {embeddings[0].shape}."
            )

    if labels is None:
        labels = [None for i in range(len(embeddings))]

    if n_jobs == 0:
        joint_embedding = copy.deepcopy(embeddings[0])
        for embedding, label in zip(embeddings[1:], labels[1:]):
            aligned_embedding = ortho_alignment.fit_transform(
                embeddings[0], embedding, labels[0], label)
            joint_embedding += aligned_embedding
        joint_embedding = joint_embedding / len(embeddings)

    else:
        delayed_func = [
            joblib.delayed(ortho_alignment.fit_transform)(embeddings[0],
                                                          embedding, labels[0],
                                                          label)
            for embedding, label in zip(embeddings[1:], labels[1:])
        ]
        parallel_pool = joblib.Parallel(n_jobs=n_jobs)
        res = parallel_pool(delayed_func)

        joint_embedding = (embeddings[0] + sum(res)) / len(embeddings)

    if post_norm:
        joint_embedding = joint_embedding / np.linalg.norm(
            joint_embedding, axis=1, keepdims=True)

    return joint_embedding



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import math

import numpy as np
import pytest
import torch

import cebra
import cebra.integrations.sklearn.cebra as cebra_sklearn_cebra
import cebra.integrations.sklearn.metrics as cebra_sklearn_metrics


def test_imports():
    import cebra

    assert hasattr(cebra, "sklearn")


def test_sklearn_infonce_loss():
    max_loss_iterations = 2
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture="offset10-model",
        max_iterations=5,
        batch_size=128,
    )

    # Example data
    X = torch.tensor(np.random.uniform(0, 1, (1000, 50)))
    y_c1 = torch.tensor(np.random.uniform(0, 1, (1000, 5)))
    y_d = np.random.randint(0, 10, (1000,))

    X_test = torch.tensor(np.random.uniform(0, 1, (500, 50)))
    X_test_2 = torch.tensor(np.random.uniform(0, 1, (600, 30)))
    y_c1_test = torch.tensor(np.random.uniform(0, 1, (500, 5)))
    y_c1_test_2 = torch.tensor(np.random.uniform(0, 1, (600, 2)))
    y_d_test = np.random.randint(0, 10, (500,))

    # Single session
    cebra_model.fit(X, y_c1)
    score = cebra.sklearn.metrics.infonce_loss(cebra_model,
                                               X_test,
                                               y_c1_test,
                                               session_id=0,
                                               num_batches=max_loss_iterations)
    assert isinstance(score, float)
    score = cebra.sklearn.metrics.infonce_loss(cebra_model,
                                               X_test,
                                               y_c1_test,
                                               num_batches=max_loss_iterations)
    assert isinstance(score, float)
    score = cebra.sklearn.metrics.infonce_loss(
        cebra_model,
        X_test,
        y_c1_test,
        num_batches=max_loss_iterations,
        correct_by_batchsize=True,
    )
    assert isinstance(score, float)

    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            X_test,
            y_c1_test,
            num_batches=max_loss_iterations,
            session_id=2,
        )
    with pytest.raises(ValueError, match="Labels.*invalid"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            X_test,
            y_d_test,
            num_batches=max_loss_iterations,
            session_id=0)
    with pytest.raises(ValueError, match="Number.*index.*invalid"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model, X_test, num_batches=max_loss_iterations)
    with pytest.raises(ValueError, match="Number.*index.*invalid"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            X_test,
            y_c1_test,
            y_d_test,
            num_batches=max_loss_iterations)
    with pytest.raises(ValueError, match="Number.*index.*invalid"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            X_test,
            y_c1_test_2,
            y_d_test,
            num_batches=max_loss_iterations)

    cebra_model.fit(X, y_d)
    score = cebra.sklearn.metrics.infonce_loss(cebra_model,
                                               X_test,
                                               y_d_test,
                                               num_batches=max_loss_iterations)
    assert isinstance(score, float)
    score = cebra.sklearn.metrics.infonce_loss(cebra_model,
                                               X_test,
                                               y_d_test,
                                               num_batches=max_loss_iterations,
                                               session_id=0)
    assert isinstance(score, float)

    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            X_test,
            y_d_test,
            num_batches=max_loss_iterations,
            session_id=2)
    with pytest.raises(ValueError, match="Labels.*invalid"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            X_test,
            y_c1_test,
            num_batches=max_loss_iterations,
            session_id=0,
        )
    with pytest.raises(ValueError, match="Number.*index.*invalid"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model, X_test, num_batches=max_loss_iterations)

    # Multisession
    cebra_model.fit([X, X], [y_c1, y_c1])
    score = cebra.sklearn.metrics.infonce_loss(cebra_model,
                                               X_test,
                                               y_c1_test,
                                               num_batches=max_loss_iterations,
                                               session_id=0)
    assert isinstance(score, float)
    score = cebra.sklearn.metrics.infonce_loss(cebra_model,
                                               X_test,
                                               y_c1_test,
                                               num_batches=max_loss_iterations,
                                               session_id=1)
    assert isinstance(score, float)

    with pytest.raises(ValueError, match="Labels.*invalid"):
        cebra.sklearn.metrics.infonce_loss(cebra_model,
                                           X_test,
                                           y_d_test,
                                           num_batches=max_loss_iterations,
                                           session_id=0)
    with pytest.raises(ValueError, match="shape"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            X_test_2,
            y_c1_test,
            num_batches=max_loss_iterations,
            session_id=0,
        )
    with pytest.raises(RuntimeError, match="No.*session_id"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model, X_test, y_c1_test, num_batches=max_loss_iterations)
    with pytest.raises(RuntimeError, match="Invalid.*session_id"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            X_test,
            y_c1_test,
            num_batches=max_loss_iterations,
            session_id=3,
        )
    with pytest.raises(ValueError, match="Number.*index.*invalid"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model, X_test, num_batches=max_loss_iterations, session_id=0)
    with pytest.raises(NotImplementedError, match="Data.*invalid"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            [X_test, X_test_2],
            y_c1_test,
            num_batches=max_loss_iterations,
            session_id=0,
        )
    with pytest.raises(NotImplementedError, match="Labels.*invalid"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model,
            X_test,
            [y_c1_test, y_c1_test],
            num_batches=max_loss_iterations,
            session_id=0,
        )

    # No batch size
    cebra_model_no_bs = cebra_sklearn_cebra.CEBRA(
        model_architecture="offset10-model",
        max_iterations=max_loss_iterations,
        batch_size=None,
    )

    cebra_model_no_bs.fit(X_test)
    score = cebra.sklearn.metrics.infonce_loss(cebra_model_no_bs,
                                               X_test,
                                               num_batches=max_loss_iterations)

    with pytest.raises(ValueError, match="Batch.*size"):
        score = cebra.sklearn.metrics.infonce_loss(
            cebra_model_no_bs,
            X_test,
            num_batches=max_loss_iterations,
            correct_by_batchsize=True,
        )


def test_sklearn_datasets_consistency():
    # Example data
    np.random.seed(42)
    embedding1 = np.random.uniform(0, 1, (10000, 4))
    embedding2 = np.random.uniform(0, 1, (10000, 10))
    embedding3 = np.random.uniform(0, 1, (8000, 6))
    embedding4 = np.random.uniform(0, 1, (5000, 7))
    embeddings_datasets = [embedding1, embedding2, embedding3, embedding4]

    labels1 = np.random.uniform(0, 1, (10000,))
    labels1_invalid = np.random.uniform(0, 1, (10000, 3))
    labels2 = np.random.uniform(0, 1, (10000,))
    labels3 = np.random.uniform(0, 1, (8000,))
    labels4 = np.random.uniform(0, 1, (5000,))
    labels_datasets = [labels1, labels2, labels3, labels4]

    dataset_ids = ["achilles", "buddy", "cicero", "gatsby"]

    # random embeddings provide R2 close to 0
    scores, pairs, datasets = cebra_sklearn_metrics.consistency_score(
        embeddings_datasets,
        dataset_ids=dataset_ids,
        labels=labels_datasets,
        between="datasets",
    )
    assert scores.shape == (12,)
    assert pairs.shape == (12, 2)
    assert len(datasets) == 4
    assert math.isclose(scores[0], 0, abs_tol=0.05)

    # no labels
    scores, pairs, datasets = cebra_sklearn_metrics.consistency_score(
        embeddings_datasets, labels=labels_datasets, between="datasets")
    assert scores.shape == (12,)
    assert pairs.shape == (12, 2)
    assert len(datasets) == 4

    # identical embeddings provide R2 close to 1
    scores, pairs, datasets = cebra_sklearn_metrics.consistency_score(
        [embedding1, embedding1],
        dataset_ids=["achilles", "buddy"],
        labels=[labels1, labels1],
        between="datasets",
    )
    assert scores.shape == (2,)
    assert pairs.shape == (2, 2)
    assert len(datasets) == 2
    assert math.isclose(scores[0], 1, abs_tol=1e-9)

    # Tensor
    scores, pairs, datasets = cebra_sklearn_metrics.consistency_score(
        [torch.Tensor(embedding) for embedding in embeddings_datasets],
        dataset_ids=dataset_ids,
        labels=[torch.Tensor(label) for label in labels_datasets],
        between="datasets",
    )
    assert scores.shape == (12,)
    assert pairs.shape == (12, 2)
    assert len(datasets) == 4

    with pytest.raises(ValueError, match="Missing.*between"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(
            embeddings_datasets, labels=labels_datasets)
    with pytest.raises(ValueError, match="Missing.*labels"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(
            embeddings_datasets,
            between="datasets",
        )
    with pytest.raises(ValueError, match="Invalid.*labels"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(
            embeddings_datasets,
            labels=[labels1, labels2],
            between="datasets",
        )
    with pytest.raises(ValueError, match="Invalid.*dataset_ids"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(
            embeddings=[embedding1, embedding4],
            labels=[labels1, labels4],
            dataset_ids=["achilles", "achilles"],
            between="datasets",
        )
    with pytest.raises(ValueError, match="Invalid.*dtype"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(
            embeddings=[embedding1, embedding2],
            labels=[
                ["test" for i in range(len(embedding1))],
                ["test2" for j in range(len(embedding2))],
            ],
            dataset_ids=["achilles", "buddy"],
            between="datasets",
        )
    with pytest.raises(NotImplementedError, match="Invalid.*labels"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(
            embeddings=[embedding1, embedding2],
            labels=[
                labels1_invalid,
                labels2,
            ],
            dataset_ids=["achilles", "buddy"],
            between="datasets",
        )


def test_sklearn_runs_consistency():
    # Example data
    np.random.seed(42)
    embedding1 = np.random.uniform(0, 1, (10000, 4))
    embedding2 = np.random.uniform(0, 1, (10000, 10))
    embedding3 = np.random.uniform(0, 1, (8000, 10))
    embeddings_runs = [embedding1, embedding2, embedding1, embedding2]
    invalid_embeddings_runs = [embedding1, embedding2, embedding3]

    labels1 = np.random.uniform(0, 1, (10000,))
    labels2 = np.random.uniform(0, 1, (10000,))
    labels_runs = [labels1, labels2, labels1, labels2]

    # between-runs consistency
    scores, pairs, ids = cebra_sklearn_metrics.consistency_score(
        embeddings=embeddings_runs, between="runs")
    assert scores.shape == (12,)
    assert pairs.shape == (12, 2)
    assert len(ids) == 4
    assert math.isclose(scores[0], 0, abs_tol=0.05)

    scores, pairs, ids = cebra_sklearn_metrics.consistency_score(
        embeddings=[embedding1, embedding1], between="runs")
    assert scores.shape == (2,)
    assert pairs.shape == (2, 2)
    assert len(ids) == 2
    assert math.isclose(scores[0], 1, abs_tol=1e-9)

    scores, pairs, ids = cebra_sklearn_metrics.consistency_score(
        [torch.Tensor(embedding) for embedding in embeddings_runs],
        between="runs",
    )
    assert scores.shape == (12,)
    assert pairs.shape == (12, 2)
    assert len(ids) == 4

    with pytest.raises(ValueError, match="No.*dataset.*ID"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(
            embeddings_runs,
            dataset_ids=["run1", "run2", "run3", "run4"],
            between="runs")

    with pytest.raises(ValueError, match="Missing.*between"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(embeddings_runs)

    with pytest.raises(ValueError, match="No.*labels"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(embeddings_runs,
                                                          labels=labels_runs,
                                                          between="runs")

    with pytest.raises(ValueError, match="Invalid.*embeddings"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(
            embeddings=[embedding1], between="runs")

    with pytest.raises(ValueError, match="Invalid.*embeddings"):
        _, _, _ = cebra_sklearn_metrics.consistency_score(
            invalid_embeddings_runs, between="runs")


@pytest.mark.parametrize("seed", [42, 24, 10])
def test_goodness_of_fit_score(seed):
    """
    Ensure that the GoF score is close to 0 for a model fit on random data.
    """
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture="offset1-model",
        max_iterations=5,
        batch_size=512,
    )
    generator = torch.Generator().manual_seed(seed)
    X = torch.rand(5000, 50, dtype=torch.float32, generator=generator)
    y = torch.rand(5000, 5, dtype=torch.float32, generator=generator)
    cebra_model.fit(X, y)
    score = cebra_sklearn_metrics.goodness_of_fit_score(cebra_model,
                                                        X,
                                                        y,
                                                        session_id=0,
                                                        num_batches=500)
    assert isinstance(score, float)
    assert np.isclose(score, 0, atol=0.01)


@pytest.mark.parametrize("seed", [42, 24, 10])
def test_goodness_of_fit_history(seed):
    """
    Ensure that the GoF score is higher for a model fit on data with underlying
    structure than for a model fit on random data.
    """

    # Generate data
    generator = torch.Generator().manual_seed(seed)
    X = torch.rand(1000, 50, dtype=torch.float32, generator=generator)
    y_random = torch.rand(len(X), 5, dtype=torch.float32, generator=generator)
    linear_map = torch.randn(50, 5, dtype=torch.float32, generator=generator)
    y_linear = X @ linear_map

    def _fit_and_get_history(X, y):
        cebra_model = cebra_sklearn_cebra.CEBRA(
            model_architecture="offset1-model",
            max_iterations=150,
            batch_size=512,
            device="cpu")
        cebra_model.fit(X, y)
        history = cebra_sklearn_metrics.goodness_of_fit_history(cebra_model)
        # NOTE(stes): Ignore the first 5 iterations, they can have nonsensical values
        # due to numerical issues.
        return history[5:]

    history_random = _fit_and_get_history(X, y_random)
    history_linear = _fit_and_get_history(X, y_linear)

    assert isinstance(history_random, np.ndarray)
    assert history_random.shape[0] > 0
    # NOTE(stes): Ignore the first 5 iterations, they can have nonsensical values
    # due to numerical issues.
    history_random_non_negative = history_random[history_random >= 0]
    np.testing.assert_allclose(history_random_non_negative, 0, atol=0.075)

    assert isinstance(history_linear, np.ndarray)
    assert history_linear.shape[0] > 0

    assert np.all(history_linear[-20:] > history_random[-20:])


@pytest.mark.parametrize("seed", [42, 24, 10])
def test_infonce_to_goodness_of_fit(seed):
    """Test the conversion from InfoNCE loss to goodness of fit metric."""
    # Test with model
    cebra_model = cebra_sklearn_cebra.CEBRA(
        model_architecture="offset10-model",
        max_iterations=5,
        batch_size=128,
    )
    generator = torch.Generator().manual_seed(seed)
    X = torch.rand(1000, 50, dtype=torch.float32, generator=generator)
    cebra_model.fit(X)

    # Test single value
    gof = cebra_sklearn_metrics.infonce_to_goodness_of_fit(1.0,
                                                           model=cebra_model)
    assert isinstance(gof, float)

    # Test array of values
    infonce_values = np.array([1.0, 2.0, 3.0])
    gof_array = cebra_sklearn_metrics.infonce_to_goodness_of_fit(
        infonce_values, model=cebra_model)
    assert isinstance(gof_array, np.ndarray)
    assert gof_array.shape == infonce_values.shape

    # Test with explicit batch_size and num_sessions
    gof = cebra_sklearn_metrics.infonce_to_goodness_of_fit(1.0,
                                                           batch_size=128,
                                                           num_sessions=1)
    assert isinstance(gof, float)

    # Test error cases
    with pytest.raises(ValueError, match="batch_size.*should not be provided"):
        cebra_sklearn_metrics.infonce_to_goodness_of_fit(1.0,
                                                         model=cebra_model,
                                                         batch_size=128)

    with pytest.raises(ValueError, match="batch_size.*should not be provided"):
        cebra_sklearn_metrics.infonce_to_goodness_of_fit(1.0,
                                                         model=cebra_model,
                                                         num_sessions=1)

    # Test with unfitted model
    unfitted_model = cebra_sklearn_cebra.CEBRA(max_iterations=5)
    with pytest.raises(RuntimeError, match="Fit the CEBRA model first"):
        cebra_sklearn_metrics.infonce_to_goodness_of_fit(1.0,
                                                         model=unfitted_model)

    # Test with model having batch_size=None
    none_batch_model = cebra_sklearn_cebra.CEBRA(batch_size=None,
                                                 max_iterations=5)
    none_batch_model.fit(X)
    with pytest.raises(ValueError, match="Computing the goodness of fit"):
        cebra_sklearn_metrics.infonce_to_goodness_of_fit(1.0,
                                                         model=none_batch_model)

    # Test missing batch_size or num_sessions when model is None
    with pytest.raises(ValueError, match="batch_size.*and num_sessions"):
        cebra_sklearn_metrics.infonce_to_goodness_of_fit(1.0, batch_size=128)

    with pytest.raises(ValueError, match="batch_size.*and num_sessions"):
        cebra_sklearn_metrics.infonce_to_goodness_of_fit(1.0, num_sessions=1)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Solvers for supervised training

Note:
    It is inclear whether these will be kept. Consider the implementation
    as experimental/outdated, and the API for this particular package unstable.
"""

import torch

import cebra.solver.base as abc_


class SupervisedNNSolver(abc_.Solver):
    """Supervised neural network training with MSE loss"""

    _variant_name = "supervised-nn"

    def fit(self,
            loader: torch.utils.data.DataLoader,
            num_steps: int,
            valid_loader=None,
            *,
            save_frequency=None,
            valid_frequency=None,
            decode: bool = False,
            logdir: str = None):
        """Train model for the specified number of steps.

        Args:
            loader: Data loader, which is an iterator over `cebra.data.Batch` instances.
                Each batch contains reference, positive and negative input samples.
            save_frequency: If not `None`, the frequency for automatically saving model checkpoints
                to `logdir`.
            logdir:  The logging directory for writing model checkpoints. The checkpoints
                can be read again using the `solver.load` function, or manually via loading the
                state dict.
        """

        self.model.train()
        step_idx = 0
        while True:
            for _, batch in enumerate(loader):
                _ = self.step(batch)
                self._log_checkpoint(num_steps, loader, valid_loader)
                step_idx += 1
                if step_idx >= num_steps:
                    break

    def step(self, batch) -> dict:
        """Perform a single gradient update.

        Args:
            batch: The input samples

        Returns:
            Dictionary containing training metrics TODO
        """
        self.optimizer.zero_grad()
        prediction = self._inference(batch)
        loss = self.criterion(prediction, batch["label"].squeeze())
        loss.backward()
        self.optimizer.step()
        self.history.append(loss.item())
        return dict(total=loss.item())

    def _inference(self, batch):
        """Compute predictions (discrete/continuous) for the batch."""
        feature, prediction = self.model(batch["neural"])
        return prediction

    def validation(self, valid_loader):
        """Deprecated since 0.0.2."""
        total_loss = 0
        for batch in valid_loader:
            prediction = self._inference(batch)
            loss = self.criterion(prediction, batch["label"].squeeze())
            total_loss += loss.item()
        return total_loss



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Integration of CEBRA with common machine learning and visualization libraries.

This package contains a growing collection of interfaces to other Python packages.
There is no clear limit (yet) of what can go into it. The current examples include
interfaces (implemented or planned) to `scikit-learn <https://scikit-learn.org/stable/>`_,
`streamlit <https://streamlit.io/>`_, `deeplabcut <http://www.mackenziemathislab.org/deeplabcut>`_,
`matplotlib <https://matplotlib.org/>`_ and `threejs <https://threejs.org/>`_ and `plotly <https://plotly.com/>`_.

Integrations can be used for data visualization, for providing easier interfaces to using CEBRA
for a particular userbase, or any other helpful function that requires a dependency to a larger
third-party package.
"""



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Solver implementations for multi-session datasetes."""

from typing import List, Optional

import torch

import cebra
import cebra.data
import cebra.models
import cebra.solver.base as abc_
from cebra.solver import register
from cebra.solver.util import Meter


@register("multi-session")
class MultiSessionSolver(abc_.Solver):
    """Multi session training, contrasting pairs of neural data."""

    _variant_name = "multi-session"

    def _mix(self, array: torch.Tensor, idx: torch.Tensor) -> torch.Tensor:
        shape = array.shape
        n, m = shape[:2]
        mixed = array.reshape(n * m, -1)[idx]
        return mixed.reshape(shape)

    def _single_model_inference(self, batch: cebra.data.Batch,
                                model: torch.nn.Module) -> cebra.data.Batch:
        """Given a single batch of input examples, computes the feature representation/embedding.

        Args:
            batch: The input data, not necessarily aligned across the batch
                dimension. This means that ``batch.index`` specifies the map
                between reference/positive samples, if not equal ``None``.
            model: The model to use for inference.

        Returns:
            Processed batch of data. While the input data might not be aligned
            across the sample dimensions, the output data should be aligned and
            ``batch.index`` should be set to ``None``.
        """
        batch.to(self.device)
        ref = torch.stack([model(batch.reference)], dim=0)
        pos = torch.stack([model(batch.positive)], dim=0)
        neg = torch.stack([model(batch.negative)], dim=0)

        pos = self._mix(pos, batch.index_reversed)

        num_features = neg.shape[2]

        return cebra.data.Batch(
            reference=ref.view(-1, num_features),
            positive=pos.view(-1, num_features),
            negative=neg.view(-1, num_features),
        )

    def _inference(self, batches: List[cebra.data.Batch]) -> cebra.data.Batch:
        """Given batches of input examples, computes the feature representations/embeddings.

        Args:
            batches: A list of input data, not necessarily aligned across the batch
                dimension. This means that ``batch.index`` specifies the map
                between reference/positive samples, if not equal ``None``.

        Returns:
            Processed batch of data. While the input data might not be aligned
            across the sample dimensions, the output data should be aligned and
            ``batch.index`` should be set to ``None``.

        """
        refs = []
        poss = []
        negs = []

        for batch, model in zip(batches, self.model):
            batch.to(self.device)
            refs.append(model(batch.reference))
            poss.append(model(batch.positive))
            negs.append(model(batch.negative))
        ref = torch.stack(refs, dim=0)
        pos = torch.stack(poss, dim=0)
        neg = torch.stack(negs, dim=0)

        pos = self._mix(pos, batches[0].index_reversed)

        num_features = neg.shape[2]

        return cebra.data.Batch(
            reference=ref.view(-1, num_features),
            positive=pos.view(-1, num_features),
            negative=neg.view(-1, num_features),
        )

    def validation(self, loader, session_id: Optional[int] = None):
        """Compute score of the model on data.

        Note:
            Overrides :py:meth:`cebra.solver.base.Solver.validation` in :py:class:`cebra.solver.base.Solver`.
        Args:
            loader: Data loader, which is an iterator over :py:class:`cebra.data.datatypes.Batch` instances.
                Each batch contains reference, positive and negative input samples.
            session_id: The session ID, an integer between 0 and the number of sessions in the
                multisession model, set to None for single session.

        Returns:
            Loss averaged over iterations on data batch.
        """

        assert session_id is not None

        iterator = self._get_loader(loader)  # loader is single session
        total_loss = Meter()
        self.model[session_id].eval()
        for _, batch in iterator:
            prediction = self._single_model_inference(batch,
                                                      self.model[session_id])
            loss, _, _ = self.criterion(prediction.reference,
                                        prediction.positive,
                                        prediction.negative)
            total_loss.add(loss.item())
        return total_loss.average


@register("multi-session-aux")
class MultiSessionAuxVariableSolver(abc_.Solver):
    """Multi session training, contrasting neural data against behavior."""

    _variant_name = "multi-session-aux"
    reference_model: torch.nn.Module

    def _inference(self, batches):
        refs = []
        poss = []
        negs = []
        for batch, model in zip(batches, self.model):
            batch.to(self.device)
            refs.append(model(batch.reference.cuda()))
            poss.append(model(batch.positive.cuda()))
            negs.append(model(batch.negative.cuda()))

        ref = torch.stack(refs, dim=0)
        pos = self._mix(torch.stack(poss, dim=0))
        neg = torch.stack(negs, dim=0)
        num_features = neg.shape[2]

        return cebra.data.Batch(
            reference=ref.view(-1, num_features),
            positive=pos.view(-1, num_features),
            negative=neg.view(-1, num_features),
        )



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Data loaders use distributions and indices to make samples available for training.

This package contains all helper functions and classes for defining and loading datasets
in the various usage modes of CEBRA, e.g. single- and multi-session datasets.
It is non-specific to a particular dataset (see :py:mod:`cebra.datasets` for actual dataset
implementations). However, the base classes for all datasets are defined here, as well as helper
functions to interact with datasets.

CEBRA supports different dataset types out-of-the box:

- :py:class:`cebra.data.single_session.SingleSessionDataset` is the abstract base class for a single session dataset. Single session datasets
  have the same feature dimension across the samples (e.g., neural data) and all context
  variables (e.g. behavior, stimuli, etc.).
- :py:class:`cebra.data.multi_session.MultiSessionDataset` is the abstract base class for a multi session dataset.
  Multi session datasets contain of multiple single session datasets. Crucially, the dimensionality of the
  auxiliary variable dimension needs to match across the sessions, which allows alignment of multiple sessions.
  The dimensionality of the signal variable can vary arbitrarily between sessions.

Note that the actual implementation of datasets (e.g. for benchmarking) is done in the :py:mod:`cebra.datasets`
package.

"""

# NOTE(stes): intentional ordering of imports to avoid circular imports
#             these imports will not be reordered by isort (see .isort.cfg)
from cebra.data.base import *
from cebra.data.datatypes import *

from cebra.data.single_session import *
from cebra.data.multi_session import *

from cebra.data.datasets import *

from cebra.data.helper import *



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
import pytest
import scipy.stats
import sklearn.model_selection

import cebra.data.helper as cebra_data_helper
import cebra.datasets


def _initialize_orthogonal_alignment_data():
    embedding_100_4d = np.random.uniform(0, 1, (100, 4))
    embedding_150_4d = np.random.uniform(0, 1, (150, 4))
    embedding_1_4d = np.random.uniform(0, 1, (1, 4))
    embedding_100_1d = np.random.uniform(0, 1, (100, 1))
    embedding_100_none = np.random.uniform(0, 1, (100,))

    labels_100_1d = np.random.uniform(0, 1, (100, 1))
    labels_150_1d = np.random.uniform(0, 1, (150, 1))
    labels_100_3d = np.random.uniform(0, 1, (100, 3))
    labels_150_3d = np.random.uniform(0, 1, (150, 3))
    labels_100_none = np.random.uniform(0, 1, (100,))
    labels_1_1d = np.random.uniform(0, 1, (1, 1))

    orthogonal_alignment_data = []

    # different sample sizes
    orthogonal_alignment_data.append(
        (embedding_100_4d, embedding_150_4d, labels_100_1d, labels_150_1d))
    # different sample sizes and more than one feature in the labels
    orthogonal_alignment_data.append(
        (embedding_100_4d, embedding_150_4d, labels_100_3d, labels_150_3d))
    # embedding to align has a single sample
    orthogonal_alignment_data.append(
        (embedding_100_4d, embedding_1_4d, labels_100_1d, labels_1_1d))
    # embeddings with one feature dimension only
    orthogonal_alignment_data.append(
        (embedding_100_1d, embedding_100_1d, labels_100_1d, labels_100_1d))
    # embeddings with one feature dimension only, as None
    orthogonal_alignment_data.append(
        (embedding_100_none, embedding_100_none, labels_100_1d, labels_100_1d))
    # labels with one feature dimension only, as None
    orthogonal_alignment_data.append(
        (embedding_100_4d, embedding_100_4d, labels_100_none, labels_100_none))
    # same sample size, no labels
    orthogonal_alignment_data.append(
        (embedding_100_4d, embedding_100_4d, None, None))

    return orthogonal_alignment_data


def _initialize_invalid_orthogonal_alignment_data():
    embedding_100_4d = np.random.uniform(0, 1, (100, 4))
    embedding_150_4d = np.random.uniform(0, 1, (150, 4))
    embedding_110_6d = np.random.uniform(0, 1, (110, 6))
    embedding_1_4d = np.random.uniform(0, 1, (1, 4))

    labels_100_1d = np.random.uniform(0, 1, (100, 1))
    labels_150_1d = np.random.uniform(0, 1, (150, 1))
    labels_110_1d = np.random.uniform(0, 1, (110, 1))
    labels_1_1d = np.random.uniform(0, 1, (1, 1))

    orthogonal_alignment_data = []

    # Embeddings with different number of features
    orthogonal_alignment_data.append(
        (embedding_100_4d, embedding_110_6d, labels_100_1d, labels_110_1d,
         "Invalid.*data"))
    # Reference embeddings and labels are not the same sample size
    orthogonal_alignment_data.append(
        (embedding_100_4d, embedding_150_4d, labels_150_1d, labels_150_1d,
         "Mismatched.*data.*labels"))
    # Embeddings and labels to align are not the same sample size
    orthogonal_alignment_data.append(
        (embedding_100_4d, embedding_150_4d, labels_100_1d, labels_100_1d,
         "Mismatched.*data.*labels"))
    # Reference embedding contain only one sample
    orthogonal_alignment_data.append(
        (embedding_1_4d, embedding_100_4d, labels_1_1d, labels_100_1d,
         "Invalid.*reference.*data"))
    # Missing reference label
    orthogonal_alignment_data.append((embedding_100_4d, embedding_150_4d, None,
                                      labels_100_1d, "Missing.*labels"))
    # Missing alignment label
    orthogonal_alignment_data.append((embedding_100_4d, embedding_150_4d,
                                      labels_100_1d, None, "Missing.*labels"))
    # No labels for embeddings that are not already aligned on time (not the same number of samples)
    orthogonal_alignment_data.append(
        (embedding_100_4d, embedding_150_4d, None, None, "Missing.*labels"))

    return orthogonal_alignment_data


def _does_shape_match(data, aligned_data):
    if len(data.shape) == 2:
        is_matching = aligned_data.shape == data.shape
    else:
        is_matching = aligned_data.shape == (data.shape[0], 1)
    return is_matching


@pytest.mark.parametrize("ref_data,data,ref_labels,labels",
                         _initialize_orthogonal_alignment_data())
def test_orthogonal_alignment_shapes(ref_data, data, ref_labels, labels):
    alignment_model = cebra_data_helper.OrthogonalProcrustesAlignment()

    alignment_model.fit(ref_data, data, ref_labels, labels)
    aligned_embedding = alignment_model.transform(data)
    assert _does_shape_match(data, aligned_embedding)

    aligned_embedding = alignment_model.fit_transform(ref_data, data,
                                                      ref_labels, labels)
    assert _does_shape_match(data, aligned_embedding)

    # Test with non-default parameters
    alignment_model = cebra_data_helper.OrthogonalProcrustesAlignment(top_k=10)

    aligned_embedding = alignment_model.fit_transform(ref_data, data,
                                                      ref_labels, labels)
    assert _does_shape_match(data, aligned_embedding), (data.shape,
                                                        aligned_embedding.shape)


@pytest.mark.parametrize("ref_data,data,ref_labels,labels,match",
                         _initialize_invalid_orthogonal_alignment_data())
def test_invalid_orthogonal_alignment(ref_data, data, ref_labels, labels,
                                      match):
    alignment_model = cebra_data_helper.OrthogonalProcrustesAlignment()

    with pytest.raises(ValueError, match=match):
        alignment_model.fit(ref_data, data, ref_labels, labels)


def test_orthogonal_alignment_without_labels():
    random_seed = 2160
    np.random.seed(random_seed)
    embedding_100_4d = np.random.uniform(0, 1, (1000, 4))
    embedding_100_4d_2 = np.random.uniform(0, 1, (1000, 4))

    alignment_model = cebra_data_helper.OrthogonalProcrustesAlignment()

    alignment_model.fit(embedding_100_4d, embedding_100_4d_2, None, None)
    aligned_embedding = alignment_model.transform(embedding_100_4d_2)

    alignment_model.fit(embedding_100_4d, embedding_100_4d_2)
    aligned_embedding_without_labels = alignment_model.transform(
        embedding_100_4d_2)

    assert np.allclose(aligned_embedding, aligned_embedding_without_labels)


@pytest.mark.parametrize("seed", [483, 425, 166, 672, 123])
def test_orthogonal_alignment(seed):
    np.random.seed(seed)
    embedding_100_4d = np.random.uniform(0, 1, (1000, 4))
    orthogonal_matrix = scipy.stats.ortho_group.rvs(dim=4, random_state=seed)
    labels_100_1d = np.random.uniform(0, 1, (1000, 1))

    alignment_model = cebra_data_helper.OrthogonalProcrustesAlignment()
    aligned_embedding = alignment_model.fit_transform(ref_data=embedding_100_4d,
                                                      data=np.dot(
                                                          embedding_100_4d,
                                                          orthogonal_matrix),
                                                      ref_label=labels_100_1d,
                                                      label=labels_100_1d)
    assert np.allclose(aligned_embedding, embedding_100_4d, atol=0.03)

    # and without labels
    aligned_embedding = alignment_model.fit_transform(ref_data=embedding_100_4d,
                                                      data=np.dot(
                                                          embedding_100_4d,
                                                          orthogonal_matrix))
    assert np.allclose(aligned_embedding, embedding_100_4d, atol=0.03)


def _initialize_embedding_ensembling_data():
    embedding_100_4d = np.random.uniform(0, 1, (100, 4))
    embedding_100_4d_2 = np.random.uniform(0, 1, (100, 4))
    embedding_100_4d_3 = np.random.uniform(0, 1, (100, 4))
    embedding_110_none = np.random.uniform(0, 1, (110,))

    labels_100_1d = np.random.uniform(0, 1, (100, 1))
    labels_100_1d_2 = np.random.uniform(0, 1, (100, 1))
    labels_100_1d_3 = np.random.uniform(0, 1, (100, 1))

    embedding_ensembling_data = []

    # Embeddings of same shape, no labels
    embedding_ensembling_data.append(
        ([embedding_100_4d, embedding_100_4d_2, embedding_100_4d_3], None, 0))
    # Embeddings with a single feature dimension (as None)
    embedding_ensembling_data.append(([embedding_110_none,
                                       embedding_110_none], None, 0))
    # Embeddings and corresponding labels to use for alignment (optional)
    embedding_ensembling_data.append(
        ([embedding_100_4d, embedding_100_4d_2, embedding_100_4d_3],
         [labels_100_1d, labels_100_1d_2, labels_100_1d_3], 0))
    # Parallel processing, with 2 jobs
    embedding_ensembling_data.append(
        ([embedding_100_4d, embedding_100_4d_2, embedding_100_4d_3], None, 2))
    # Parallel processing, with all jobs available
    embedding_ensembling_data.append(
        ([embedding_100_4d, embedding_100_4d_2, embedding_100_4d_3], None, -1))

    return embedding_ensembling_data


def _initialize_invalid_embedding_ensembling_data():
    embedding_100_4d = np.random.uniform(0, 1, (100, 4))
    embedding_100_4d_2 = np.random.uniform(0, 1, (100, 4))
    embedding_100_4d_3 = np.random.uniform(0, 1, (100, 4))
    embedding_150_4d = np.random.uniform(0, 1, (150, 4))
    embedding_100_6d = np.random.uniform(0, 1, (100, 6))
    embedding_1_4d = np.random.uniform(0, 1, (1, 4))

    labels_100_1d = np.random.uniform(0, 1, (100, 1))
    labels_100_1d_2 = np.random.uniform(0, 1, (100, 1))
    labels_150_1d = np.random.uniform(0, 1, (150, 1))

    embedding_ensembling_data = []

    # Number of jobs set to None
    embedding_ensembling_data.append(
        ([embedding_100_4d, embedding_100_4d_2,
          embedding_100_4d_3], None, None, "Invalid.*n_jobs"))
    # Embeddings with different sample sizes
    embedding_ensembling_data.append(
        ([embedding_100_4d, embedding_100_4d_2,
          embedding_150_4d], None, 0, "Inconsistent.*embeddings"))
    # Embeddings with different number of features
    embedding_ensembling_data.append(
        ([embedding_100_4d, embedding_100_4d_2,
          embedding_100_6d], None, 0, "Inconsistent.*embeddings"))
    # Embeddings and labels with labels with a different sample size as their corresponding embedding
    embedding_ensembling_data.append(
        ([embedding_100_4d, embedding_100_4d_2,
          embedding_100_4d_3], [labels_100_1d, labels_100_1d_2,
                                labels_150_1d], 0, "Mismatched.*data.*labels"))
    # Embedding with a single sample (need at least {k_top})
    embedding_ensembling_data.append(
        ([embedding_1_4d, embedding_1_4d], None, 0, "Invalid.*data"))

    return embedding_ensembling_data


@pytest.mark.parametrize("embeddings,labels,n_jobs",
                         _initialize_embedding_ensembling_data())
def test_embedding_ensembling_shapes(embeddings, labels, n_jobs):

    joint_embedding = cebra_data_helper.ensemble_embeddings(
        embeddings=embeddings, labels=labels, n_jobs=n_jobs)
    assert _does_shape_match(embeddings[0], joint_embedding)


def test_embeddings_ensembling_without_labels():
    random_seed = 2160
    np.random.seed(random_seed)

    embedding_100_4d = np.random.uniform(0, 1, (100, 4))
    embedding_100_4d_2 = np.random.uniform(0, 1, (100, 4))

    # same sample size, no labels
    joint_embedding = cebra_data_helper.ensemble_embeddings(
        embeddings=[embedding_100_4d, embedding_100_4d_2], labels=[None, None])
    joint_embedding_without_labels = cebra_data_helper.ensemble_embeddings(
        embeddings=[embedding_100_4d, embedding_100_4d_2])
    assert np.allclose(joint_embedding, joint_embedding_without_labels)


@pytest.mark.parametrize("embeddings,labels,n_jobs,match",
                         _initialize_invalid_embedding_ensembling_data())
def test_invalid_embedding_ensembling(embeddings, labels, n_jobs, match):
    with pytest.raises(ValueError, match=match):
        _ = cebra_data_helper.ensemble_embeddings(
            embeddings=embeddings,
            labels=labels,
            n_jobs=n_jobs,
        )


@pytest.mark.parametrize("seed", [483, 426, 166, 674, 123])
def test_embedding_ensembling(seed):
    np.random.seed(seed)
    embedding_100_4d = np.random.uniform(0, 1, (100, 4))
    labels_100_1d = np.random.uniform(0, 1, (100, 1))
    orthogonal_matrix = scipy.stats.ortho_group.rvs(dim=4, random_state=seed)
    orthogonal_matrix_2 = scipy.stats.ortho_group.rvs(dim=4,
                                                      random_state=seed + 1)

    embedding_100_4d_2 = np.dot(embedding_100_4d, orthogonal_matrix)
    embedding_100_4d_3 = np.dot(embedding_100_4d, orthogonal_matrix_2)

    labels = [labels_100_1d, labels_100_1d, labels_100_1d]

    joint_embedding = cebra_data_helper.ensemble_embeddings(
        embeddings=[embedding_100_4d, embedding_100_4d_2, embedding_100_4d_3],
        labels=labels)
    assert np.allclose(joint_embedding, embedding_100_4d, atol=0.05)

    joint_embedding = cebra_data_helper.ensemble_embeddings(
        embeddings=[embedding_100_4d, embedding_100_4d_2, embedding_100_4d_3])
    assert np.allclose(joint_embedding, embedding_100_4d, atol=0.05)


@pytest.mark.benchmark
@pytest.mark.requires_dataset
def test_ensembling_performances(n_models=3):
    dataset = cebra.datasets.init("rat-hippocampus-single-achilles")

    scores, embeddings = [], []
    for i in range(n_models):
        cebra_model = cebra.CEBRA(model_architecture="offset10-model",
                                  max_iterations=100,
                                  batch_size=512,
                                  verbose=False)
        cebra_model.fit(dataset.neural)
        embedding = cebra_model.transform(dataset.neural)

        (train_embedding, valid_embedding, train_label,
         valid_label) = sklearn.model_selection.train_test_split(
             embedding, dataset.continuous_index[:, 0], test_size=0.3)

        decoder = cebra.KNNDecoder()
        decoder.fit(train_embedding, train_label)
        score = decoder.score(valid_embedding, valid_label)

        scores.append(score)
        embeddings.append(embedding)

    joint_embedding = cebra_data_helper.ensemble_embeddings(
        embeddings=embeddings)

    (train_embedding, valid_embedding, train_label,
     valid_label) = sklearn.model_selection.train_test_split(
         joint_embedding, dataset.continuous_index[:, 0], test_size=0.3)

    decoder = cebra.KNNDecoder()
    decoder.fit(train_embedding, train_label)
    score = decoder.score(valid_embedding, valid_label)

    assert all(score_i < score for score_i in scores)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Discrete indices."""

from typing import Literal, Optional, Union

import numpy as np
import numpy.typing as npt
import scipy.interpolate
import torch

import cebra.distributions.base as abc_
import cebra.helper


class Discrete(abc_.ConditionalDistribution, abc_.HasGenerator):
    """Resample 1-dimensional discrete data.

    The distribution is fully specified by an array of discrete samples.
    Samples can be drawn either from the dataset directly (i.e., output
    samples will have the same distribution of class labels as the samples
    used to specify the distribution), or from a resampled data distribution
    where the occurrence of each class label is balanced.

    Args:
        samples: Discrete index used for sampling
    """

    def _to_numpy_int(self, samples: Union[torch.Tensor,
                                           npt.NDArray]) -> npt.NDArray:
        if isinstance(samples, torch.Tensor):
            samples = samples.cpu().numpy()
        if not cebra.helper._is_integer(samples):
            samples = samples.astype(int)
        return samples

    def __init__(
        self,
        samples: torch.Tensor,
        device: Literal["cpu", "cuda"] = "cpu",
        seed: Optional[int] = None,
    ):
        abc_.HasGenerator.__init__(self, device=device, seed=None)
        self._set_data(samples)
        self.sorted_idx = torch.from_numpy(np.argsort(self.samples))
        self._init_transform()

    def _set_data(self, samples: torch.Tensor):
        samples = self._to_numpy_int(samples)
        if samples.ndim > 1:
            raise ValueError(
                f"Data dimensionality is {samples.shape}, but can only accept a single dimension."
            )
        self.samples = samples

    @property
    def num_samples(self) -> int:
        """Number of samples in the index."""
        return len(self.samples)

    def _init_transform(self):
        self.counts = np.bincount(self.samples)
        self.cdf = np.zeros((len(self.counts) + 1,))
        self.cdf[1:] = np.cumsum(self.counts)
        # NOTE(stes): This is the only use of a scipy function in the entire code
        # base for now. Replacing scipy.interpolate.interp1d with an equivalent
        # function from torch would make it possible to drop scipy as a dependency
        # of the package.
        self.transform = scipy.interpolate.interp1d(
            np.linspace(0, self.num_samples, len(self.cdf)), self.cdf)

    def sample_uniform(self, num_samples: int) -> torch.Tensor:
        """Draw samples from the uniform distribution over values.

        This will change the likelihood of values depending on the values
        in the given (discrete) index. When reindexing the dataset with
        the returned indices, all values in the index will appear with
        equal probability.

        Args:
            num_samples: Number of uniform samples to be drawn.

        Returns:
            A batch of indices from the distribution. Reindexing the
            index samples of this instance with the returned in indices
            will yield a uniform distribution across the discrete values.
        """
        samples = np.random.uniform(0, self.num_samples, (num_samples,))
        samples = self.transform(samples).astype(int)
        return self.sorted_idx[samples]

    def sample_empirical(self, num_samples: int) -> torch.Tensor:
        """Draw samples from the empirical distribution.

        Args:
            num_samples: Number of samples to be drawn.

        Returns:
            A batch of indices from the empirical distribution,
            which is the uniform distribution over ``[0, N-1]``.
        """
        samples = np.random.randint(0, self.num_samples, (num_samples,))
        return self.sorted_idx[samples]

    def sample_conditional(self, reference_index: torch.Tensor) -> torch.Tensor:
        """Draw samples conditional on template samples.

        Args:
            samples: batch of indices, typically drawn from a
                prior distribution. Conditional samples will match
                the values of these indices

        Returns:
            batch of indices, whose values match the values
            corresponding to the given indices.
        """
        reference_index = self._to_numpy_int(reference_index)
        idx = np.random.uniform(0, 1, len(reference_index))
        idx *= self.cdf[reference_index + 1] - self.cdf[reference_index]
        idx += self.cdf[reference_index]
        idx = idx.astype(int)

        return self.sorted_idx[idx]


class DiscreteUniform(Discrete, abc_.PriorDistribution):
    """Re-sample the given indices and produce samples from a uniform distribution."""

    def sample_prior(self, num_samples: int) -> torch.Tensor:
        return self.sample_uniform(num_samples)


class DiscreteEmpirical(Discrete, abc_.PriorDistribution):
    """Draw samples from the empirical distribution defined by the passed index."""

    def sample_prior(self, num_samples: int) -> torch.Tensor:
        return self.sample_empirical(num_samples)


DiscreteUniform.sample_prior.__doc__ = Discrete.sample_uniform.__doc__
DiscreteEmpirical.sample_prior.__doc__ = Discrete.sample_empirical.__doc__



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, Tuple, Union

import numpy as np
import numpy.typing as npt
import torch


def _get_min_max(
        labels: List[Union[npt.NDArray, torch.Tensor]]) -> Tuple[float, float]:
    """Get the min and max values in the list of labels.

    Args:
        labels: A list of arrays corresponding to the labels associated to each embeddings.

    Returns:
        The minimum value (first returns) and the maximum value (second returns).

    """
    min = float("inf")
    max = float("-inf")
    for label in labels:
        if any(isinstance(label_element, str) for label_element in label):
            raise ValueError(
                "Invalid labels dtype, expect floats or integers, got string")
        min = np.min(label) if min > np.min(label) else min
        max = np.max(label) if max < np.max(label) else max
    return min, max


def _mean_bin(data: npt.NDArray, mask: npt.NDArray) -> npt.NDArray:
    """Compute the mean across samples in `data` for the given mask.

    Args:
        data: Array of data to compute mean over.
        mask: Boolean array indicating which elements to include.

    Returns:
        Mean of the data across samples over the specified mask.
    """
    if mask.sum() == 0:
        return None
    else:
        return data[mask].mean(axis=0)


def _coarse_to_fine(
    data: Union[npt.NDArray, torch.Tensor],
    digitized_labels: Union[npt.NDArray, torch.Tensor],
    bin_idx: int,
) -> npt.NDArray:
    """Compute a quantized feature from the data using the given digitized labels.

    Args:
        data: Array of data to compute quantized feature over.
        digitized: Array of digitized labels to find the indexes in the data for quantization.
        bin_idx: Integer to which values from the labels that are close enough to are grouped
            to be averaged.

    Returns:
        Quantized data for the specified bin.
    """
    for difference in [0, 1, 2]:
        mask = abs(digitized_labels - bin_idx) <= difference
        quantized_data = _mean_bin(data, mask)
        if quantized_data is not None:
            return quantized_data
    raise ValueError(
        f"Digitalized labels does not have elements close enough to bin index {bin_idx}. "
        f"The bin index should be in the range of the labels values.")


def align_embeddings(
    embeddings: List[Union[npt.NDArray, torch.Tensor]],
    labels: List[Union[npt.NDArray, torch.Tensor]],
    normalize: bool = True,
    n_bins: int = 100,
) -> List[Union[npt.NDArray, torch.Tensor]]:
    """Align the embeddings in the ``embeddings`` list to the ``labels``.

    Each embedding has an associated set of labels. During alignment, the labels are
    digitalized so that all sets of digitalized labels contain the same set of values.
    Then the embeddings are quantized based on the new digitalized labels.

    Args:
        embeddings: List of embeddings to align on the labels.
        labels: List of labels corresponding to each embedding and to use for alignment
            between them.
        normalize: If True, samples of the embeddings are normalized across dimensions.
        n_bins: Number of values for the digitalized common labels.

    Returns:
        The embeddings aligned on the labels.
    """
    if n_bins < 1:
        raise ValueError(
            f"Invalid value for n_bins, expect a value at least equals to 1, got {n_bins}."
        )

    quantized_embeddings = []

    # get min and max label values to set the digitalization bins
    min_labels_value, max_labels_value = _get_min_max(labels)

    for embedding, label in zip(embeddings, labels):
        if len(embedding) != len(label):
            raise ValueError(
                "Labels invalid, the labels associated to an embedding should "
                f"contain the same number of samples as its embedding, got "
                f"{len(embeddings)} samples in embedding and {len(label)} samples in label"
            )
        if isinstance(embedding, torch.Tensor):
            embedding = embedding.numpy()
        if isinstance(label, torch.Tensor):
            label = label.numpy()

        # Only keep non-NaNs values
        valid_ = np.isnan(embedding).any(axis=1)
        valid_embedding = embedding[~valid_]
        valid_labels = label[~valid_]

        # Digitize the labels so that values are contained into a common set of values
        digitized_labels = np.digitize(
            valid_labels, np.linspace(min_labels_value, max_labels_value,
                                      n_bins))

        # quantize embedding based on the new labels
        quantized_embedding = [
            _coarse_to_fine(valid_embedding, digitized_labels, bin_idx)
            for bin_idx in range(n_bins)[1:]
        ]

        if normalize:  # normalize across dimensions
            quantized_embedding_norm = [
                quantized_sample / np.linalg.norm(quantized_sample, axis=0)
                for quantized_sample in quantized_embedding
            ]

        quantized_embeddings.append(quantized_embedding_norm)
    return quantized_embeddings



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Plotly interface to CEBRA."""
from typing import Optional, Tuple, Union

import matplotlib.cm
import matplotlib.colors
import numpy as np
import numpy.typing as npt
import plotly.graph_objects
import torch

from cebra.integrations.matplotlib import _EmbeddingPlot


def _convert_cmap2colorscale(cmap: str, pl_entries: int = 11, rdigits: int = 2):
    """Convert matplotlib colormap to plotly colorscale.

    Args:
        cmap: A registered colormap name from matplotlib.
        pl_entries: Number of colors to use in the plotly colorscale.
        rdigits: Number of digits to round the colorscale to.

    Returns:
        pl_colorscale: List of scaled colors to plot the embeddings
    """
    scale = np.linspace(0, 1, pl_entries)
    colors = (cmap(scale)[:, :3] * 255).astype(np.uint8)
    pl_colorscale = [[float(round(s, rdigits)), f"rgb{tuple(color.tolist())}"]
                     for s, color in zip(scale, colors)]
    return pl_colorscale


class _EmbeddingInteractivePlot(_EmbeddingPlot):

    def __init__(self, **kwargs):
        self.figsize = kwargs.get("figsize")
        super().__init__(**kwargs)
        self.colorscale = self._define_colorscale(self.cmap)

    def _define_ax(self, axis: Optional[plotly.graph_objects.Figure]):
        """Define the axis of the plot.

        Args:
            axis: Optional axis to create the plot on.

        Returns:
            axis: The axis :py:meth:`plotly.graph_objs._figure.Figure` of the plot.
        """

        if axis is None:
            self.axis = plotly.graph_objects.Figure(
                layout=plotly.graph_objects.Layout(height=100 * self.figsize[0],
                                                   width=100 * self.figsize[1]))

        else:
            self.axis = axis

    def _define_colorscale(self, cmap: str):
        """Specify the cmap for plotting the latent space.

        Args:
            cmap: The Colormap instance or registered colormap name used to map scalar data to colors.
                It will be ignored if `embedding_labels` is set to a valid RGB(A).


        Returns:
            colorscale: List of scaled colors to plot the embeddings
        """
        colorscale = _convert_cmap2colorscale(matplotlib.cm.get_cmap(cmap))

        return colorscale

    def _plot_3d(self, **kwargs) -> plotly.graph_objects.Figure:
        """Plot the embedding in 3d.

        Returns:
            The axis :py:meth:`plotly.graph_objs._figure.Figure` of the plot.
        """
        showlegend = kwargs.get("showlegend", False)
        discrete = kwargs.get("discrete", False)
        col = kwargs.get("col", None)
        row = kwargs.get("row", None)
        template = kwargs.get("template", "plotly_white")
        data = []

        if not discrete and showlegend:
            raise ValueError("Cannot show legend with continuous labels.")

        idx1, idx2, idx3 = self.idx_order

        if discrete:
            unique_labels = np.unique(self.embedding_labels)
        else:
            unique_labels = [self.embedding_labels]

        for label in unique_labels:
            if discrete:
                filtered_idx = [
                    i for i, x in enumerate(self.embedding_labels) if x == label
                ]
            else:
                filtered_idx = np.arange(self.embedding.shape[0])
            data.append(
                plotly.graph_objects.Scatter3d(x=self.embedding[filtered_idx,
                                                                idx1],
                                               y=self.embedding[filtered_idx,
                                                                idx2],
                                               z=self.embedding[filtered_idx,
                                                                idx3],
                                               mode="markers",
                                               marker=dict(
                                                   size=self.markersize,
                                                   opacity=self.alpha,
                                                   color=label,
                                                   colorscale=self.colorscale,
                                               ),
                                               name=str(label)))

        for trace in data:
            if col is None or row is None:
                self.axis.add_trace(trace)
            else:
                self.axis.add_trace(trace, row=row, col=col)

        self.axis.update_layout(
            template=template,
            showlegend=showlegend,
            title=self.title,
        )

        return self.axis


def plot_embedding_interactive(
    embedding: Union[npt.NDArray, torch.Tensor],
    embedding_labels: Optional[Union[npt.NDArray, torch.Tensor, str]] = "grey",
    axis: Optional[plotly.graph_objects.Figure] = None,
    markersize: float = 1,
    idx_order: Optional[Tuple[int]] = None,
    alpha: float = 0.4,
    cmap: str = "cool",
    title: str = "Embedding",
    figsize: Tuple[int] = (5, 5),
    dpi: int = 100,
    **kwargs,
) -> plotly.graph_objects.Figure:
    """Plot embedding in a 3D dimensional space.

    This is supposing that the dimensions provided to ``idx_order`` are in the range of the number of
    dimensions of the embedding (i.e., between 0 and :py:attr:`cebra.CEBRA.output_dimension` -1).

    The function makes use of :py:class:`plotly.graph_objects.Scatter` and parameters from that function can be provided
    as part of ``kwargs``.

    Args:
        embedding: A matrix containing the feature representation computed with CEBRA.
        embedding_labels: The labels used to map the data to color. It can be:

            - A vector that is the same sample size as the embedding, associating a value
              to each of the sample, either discrete or continuous.
            - A string, either `time`, then the labels will color the embedding based on
              temporality, or a string that can be interpreted as a RGB(A) color, then
              the embedding will be uniformly displayed with that unique color.

        axis: Optional axis to create the plot on.
        idx_order: A tuple (x, y, z) or (x, y) that maps a dimension in the data to a dimension
            in the 3D/2D embedding. The simplest form is (0, 1, 2) or (0, 1) but one might want
            to plot either those dimensions differently (e.g., (1, 0, 2)) or other dimensions
            from the feature representation (e.g., (2, 4, 5)).

        markersize: The marker size.
        alpha: The marker blending, between 0 (transparent) and 1 (opaque).
        cmap: The Colormap instance or registered colormap name used to map scalar data to colors.
            It will be ignored if `embedding_labels` is set to a valid RGB(A).
        title: The title on top of the embedding.
        figsize: Figure width and height in inches.
        dpi: Figure resolution.
        kwargs: Optional arguments to customize the plots. This dictionary includes the following
            optional arguments:

            - ``showlegend``: Whether to show the legend or not.
            - ``discrete``: Whether the labels are discrete or not.
            - ``col``: The column of the subplot to plot the embedding on.
            - ``row``: The row of the subplot to plot the embedding on.
            - ``template``: The template to use for the plot.

            Note: ``showlegend`` can be ``True`` only if ``discrete`` is ``True``.
            See :py:class:`plotly.graph_objects.Scatter` documentation for more details on which
            arguments to use.

    Example:

        >>> import cebra
        >>> import numpy as np
        >>> X = np.random.uniform(0, 1, (100, 50))
        >>> y = np.random.uniform(0, 10, (100, 5))
        >>> cebra_model = cebra.CEBRA(max_iterations=10)
        >>> cebra_model.fit(X, y)
        CEBRA(max_iterations=10)
        >>> embedding = cebra_model.transform(X)
        >>> cebra_time = np.arange(X.shape[0])
        >>> fig = cebra.plot_embedding_interactive(embedding, embedding_labels=cebra_time)

    """
    return _EmbeddingInteractivePlot(
        embedding=embedding,
        embedding_labels=embedding_labels,
        axis=axis,
        idx_order=idx_order,
        markersize=markersize,
        alpha=alpha,
        cmap=cmap,
        title=title,
        figsize=figsize,
        dpi=dpi,
    ).plot(**kwargs)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utilities for performing a grid search across CEBRA models."""

import copy
import pathlib
import pickle
from typing import Iterable, List, Literal, Optional, Tuple

import matplotlib.axes
import numpy as np
import pandas as pd
import sklearn.model_selection
import sklearn.utils.validation as sklearn_utils_validation
import torch

import cebra.integrations.matplotlib as cebra_matplotlib
import cebra.integrations.sklearn.cebra as cebra_sklearn_cebra


class GridSearch:
    """Define and run a grid search on the CEBRA hyperparameters.

    Note:
        We recommend that you use that grid-search implementation for rather small and simple grid-search.

    Depending on the usage, one needs to optimize some parameters used in the CEBRA model, e.g., the
    :py:attr:`~.CEBRA.temperature`, the :py:attr:`~.CEBRA.batch_size`, the :py:attr:`~.CEBRA.learning_rate`.
    A grid-search on that set of parameters consists in finding the best combination of values
    for those parameters. For that, models with different combinations of parameters are trained and the
    parameters used to get the best performing model are considered to be the optimal parameters. One can also
    define the fixed parameters, which will stay constant
    from one model to the other, e.g., the :py:attr:`~.CEBRA.max_iterations` or the :py:attr:`~.CEBRA.verbose`.

    The class also allows to iterate over multiple datasets and combinations of auxiliary variables.
    """

    def generate_models(self, params: dict) -> Tuple[dict, List[dict]]:
        """Generate the models to compare, based on fixed and variable CEBRA parameters.

        Args:
            params: Dict of parameter values provided by the user, either as a single value, for
                fixed hyperparameter values, or with a list of values for hyperparameters to optimize.
                If the value is a list of a single element, the hyperparameter is considered as fixed.

        Returns:
            A dict of (unfitted) models (first returns) and a list of dict of the parameters for each
            model (second returns).

        Example:

            >>> import cebra.grid_search
            >>> # 1. Define the parameters for the models
            >>> params_grid = dict(
            ...     output_dimension = [3, 16],
            ...     learning_rate = [0.001],
            ...     time_offsets = 5,
            ...     max_iterations = 10,
            ...     verbose = False)
            >>> # 2. Define the grid search and generate the models
            >>> grid_search = cebra.grid_search.GridSearch()
            >>> models, parameter_grid = grid_search.generate_models(params=params_grid)

        """
        # determine if each param is fixed or to optimize
        fixed_params, variable_params = {}, {}
        for key in list(params.keys()):
            if not isinstance(params[key], list):  # fixed parameters
                if (isinstance(params[key], dict) or
                        isinstance(params[key], tuple) or
                        isinstance(params[key], torch.Tensor) or
                        isinstance(params[key], np.ndarray)):
                    raise ValueError(
                        f"Invalid list of variable parameters, provide a list of values, got {type(params[key])}."
                    )
                fixed_params[key] = params[key]
            elif len(params[key]) < 2:  # fixed parameters
                if not isinstance(params[key][0], Iterable) or isinstance(
                        params[key][0], str):
                    fixed_params[key] = params[key][0]
            elif not any(
                    isinstance(params[key][i], Iterable) or
                    isinstance(params[key][0], str)
                    for i in range(len(params[key]))):  # variable parameters
                variable_params[key] = params[key]
            else:
                raise ValueError(
                    f"Invalid parameter {params[key]} of type {type(params[key])}."
                )

        parameter_grid = list(
            sklearn.model_selection.ParameterGrid(variable_params))

        models = {}
        for params in parameter_grid:
            params_name = [f"{name}_{value}" for name, value in params.items()]
            model_name = "_".join(map(str, list(params_name)))
            models[model_name] = cebra_sklearn_cebra.CEBRA(
                **params, **fixed_params)

        return models, parameter_grid

    def fit_models(self,
                   datasets: dict,
                   params: dict,
                   models_dir: str = "saved_models") -> "GridSearch":
        """Fit the models to compare in the grid search on the provided datasets.

        Args:
            datasets: A dict of datasets to iterate over. The values in the dict can either be a tuple or an
                iterable structure (``Iterable``, i.e., list, :py:func:`numpy.array`, :py:class:`torch.Tensor`).
                If the value provided for a given dataset is a tuple, contain multiple elements and those elements are all iterable structures (as defined
                before), then the first element is considered to be the ``data`` to fit the CEBRA models (``X``
                to be used in :py:meth:`cebra.CEBRA.fit`) on and the other values to be the auxiliary variables
                be to use in the training process (``y`` s to be used in :py:meth:`cebra.CEBRA.fit`). The models
                are then trained using behavioral contrastive learning (either CEBRA-Behavior or CEBRA-Hybrid).
                If the value provided for a given dataset is a tuple containing a single value and it is an iterable structure (as defined before) or is
                such an iterable structure directly (not a tuple), then the value is considered to be the ``data``
                to fit the CEBRA models on. The models are then trained using temporal contrastive learning
                (CEBRA-Time).
                An example of a valid ``datasets`` value could be:
                ``datasets={"dataset1": neural_data, "dataset2": (neurald_data, continuous_data, discrete_data),
                "dataset3": (neural_data2, continuous_data2)}``.
            params: Dict of parameter values provided by the user, either as a single value, for
                fixed hyperparameter values, or with a list of values for hyperparameters to optimize.
                If the value is a list of a single element, the hyperparameter is considered as fixed.
            models_dir: The path to the folder in which save the (fitted) models.

        Returns:
            ``self`` for chaining operations.

        Example:

            >>> import cebra.grid_search
            >>> import numpy as np
            >>> neural_data =  np.random.uniform(0, 1, (300, 30))
            >>> # 1. Define the parameters for the models
            >>> params_grid = dict(
            ...     output_dimension = [3, 16],
            ...     learning_rate = [0.001],
            ...     time_offsets = 5,
            ...     max_iterations = 10,
            ...     verbose = False)
            >>> # 2. Fit the models generated from the list of parameters
            >>> grid_search = cebra.grid_search.GridSearch()
            >>> grid_search = grid_search.fit_models(datasets={"neural_data": neural_data},
            ...                                      params=params_grid,
            ...                                      models_dir="grid_search_models")

        """
        if models_dir is not None:
            models_dir = pathlib.Path(models_dir)
            if not pathlib.Path.exists(models_dir):
                pathlib.Path.mkdir(models_dir)
        self.models_dir = models_dir

        # get the list of models to iterate over based on the parameters
        unfitted_models, parameter_grid = self.generate_models(params)

        self.models = {}
        self.parameter_grid = []
        self.models_names = []
        for dataset_name in datasets.keys():
            if isinstance(datasets[dataset_name], tuple):
                for data in datasets[
                        dataset_name]:  # check that data is iterable
                    if not isinstance(data, Iterable):
                        raise ValueError(
                            f"Invalid dataset {dataset_name}, provide either the data or a tuple containing the data and auxiliary variables."
                        )
                X = datasets[dataset_name][0]
                if len(datasets[dataset_name]) > 1:
                    y = datasets[dataset_name][1:]
            elif (isinstance(datasets[dataset_name], np.ndarray) or
                  isinstance(datasets[dataset_name], torch.Tensor) or
                  isinstance(datasets[dataset_name], list)):
                X = datasets[dataset_name]
                y = None
            else:
                ValueError(
                    f"Invalid dataset {dataset_name}, provide either the data or a tuple containing the data and auxiliary variables."
                )

            # fit and save each model
            for model_idx, model_name in enumerate(unfitted_models.keys()):
                model = copy.deepcopy(unfitted_models[model_name])

                if y is None:
                    model.fit(X)
                else:
                    model.fit(X, *y)

                if self.models_dir is not None:
                    model.save(self.models_dir /
                               f"{model_name}_{dataset_name}.pt")

                self.models[f"{model_name}_{dataset_name}"] = model
                self.parameter_grid.append(parameter_grid[model_idx])
                self.models_names.append(f"{model_name}_{dataset_name}")

        with open(models_dir / "parameter_grid.pkl", "wb") as fp:
            pickle.dump(self.parameter_grid, fp)
        with open(models_dir / "models_names.pkl", "wb") as fp:
            pickle.dump(self.models_names, fp)

        return self

    @classmethod
    def load(cls, dir: str) -> Tuple[cebra_sklearn_cebra.CEBRA, List[dict]]:
        """Load the *fitted* models and parameter grid present in ``dir``.

        Note:
            It is recommended to generate the models to iterate over by using :py:meth:`fit_models`, but you can also run
            the following function on a folder containing valid *fitted* models **and** the corresponding parameter grid.

        Args:
            dir: The directory in which the fitted models are saved.

        Returns:
            A dict containing the fitted models (first returns) and a list of dict containing the parameters
            used for each model present in the ``dir`` (second returns).

        Example:

            >>> import cebra.grid_search
            >>> models, parameter_grid = cebra.grid_search.GridSearch().load(dir="grid_search_models")

        """
        dir = pathlib.Path(dir)
        if not pathlib.Path.exists(dir):
            raise ValueError(
                f"Invalid model directory, provide a directory that exists and contains the models to load, got {dir}."
            )

        cosine_files = []
        euclidean_files = []
        parameter_grid = None
        models_names = None
        for filename in pathlib.Path.iterdir(dir):
            if (str(filename).endswith(".pth") or str(filename).endswith(".pt")
               ) and not str(filename).startswith(".DS_Store"):
                if "cosine" in str(filename):
                    cosine_files.append(pathlib.Path(filename))
                else:
                    euclidean_files.append(pathlib.Path(filename))
            if str(filename).endswith("parameter_grid.pkl") and not str(
                    filename).startswith(".DS_Store"):
                with open(filename, "rb") as fp:
                    parameter_grid = pickle.load(fp)
            if str(filename).endswith("models_names.pkl") and not str(
                    filename).startswith(".DS_Store"):
                with open(filename, "rb") as fp:
                    models_names = pickle.load(fp)

        all_files = sorted(cosine_files) + sorted(euclidean_files)

        if len(all_files) < 0:
            raise ValueError(
                f"Missing models, the {dir} directory doesn't contain any model."
            )
        if not parameter_grid:
            raise ValueError(
                f"Missing parameter grid, the {dir} directory doesn't contain a parameter grid file."
            )
        if not models_names:
            raise ValueError(
                f"Missing parameter grid, the {dir} directory doesn't contain a parameter grid file."
            )

        # check that all models are fitted
        models = {}
        datasets = []
        for filename in all_files:
            model = cebra_sklearn_cebra.CEBRA.load(filename)
            sklearn_utils_validation.check_is_fitted(model, "n_features_")
            if filename.stem.split("_")[-1] not in datasets:
                datasets.append(filename.stem.split("_")[-1])
            models[filename.stem] = model

        if len(models) != len(models_names) or len(models) != len(
                parameter_grid):
            raise ValueError(
                f"Number of models doesn't correspond to the number of sets of parameters in the parameter grid, "
                f"got {len(models)} against {len(models_names)}.")

        models = {k: models[k] for k in models_names}
        return models, parameter_grid

    def get_best_model(
        self,
        scoring: Literal["infonce_loss"] = "infonce_loss",
        dataset_name: Optional[str] = None,
        models_dir: str = None,
    ) -> Tuple[cebra_sklearn_cebra.CEBRA, str]:
        """Get the model with the best performance across all sets of parameters and datasets.

        Args:
            scoring: Metric to use to evaluate the models performances.
            dataset_name: Name of the dataset to find the best model for. By default, ``dataset_name``
                is set to ``None`` and the best model is searched from the list of all models for all
                sets of parameters and across all datasets. A ``dataset_name`` is valid if models were
                fitted on that set of data and those models are present in ``models_dir``. Then, the
                returned model will correspond to the model with the highest performances on *that*
                dataset only.
            models_dir: The path to the folder in which save the (fitted) models.

        Returns:
            The :py:class:`cebra.CEBRA` model with the highest performance for a given ``dataset_name``
            (first returns) and its name (second returns).

        Example:

            >>> import cebra.grid_search
            >>> import numpy as np
            >>> neural_data =  np.random.uniform(0, 1, (300, 30))
            >>> # 1. Define the parameters for the models
            >>> params_grid = dict(
            ...     output_dimension = [3, 16],
            ...     learning_rate = [0.001],
            ...     time_offsets = 5,
            ...     max_iterations = 10,
            ...     verbose = False)
            >>> # 2. Fit the models generated from the list of parameters
            >>> grid_search = cebra.grid_search.GridSearch()
            >>> grid_search = grid_search.fit_models(datasets={"neural_data": neural_data},
            ...                        params=params_grid,
            ...                        models_dir="grid_search_models")
            >>> # 3. Get model with the best performances and use it as usual
            >>> best_model, best_model_name = grid_search.get_best_model()
            >>> embedding = best_model.transform(neural_data)

        """

        if not hasattr(self, "models_dir"):
            if models_dir is None:
                raise ValueError(
                    "Missing models directory, provide a value for the models_dir parameter."
                )
            self.models_dir = models_dir

        if not hasattr(self, "models"):
            self.models, _ = GridSearch().load(self.models_dir)

        best_score = [cebra_sklearn_cebra.CEBRA(), float("inf"), "test"]
        if scoring == "infonce_loss":
            for model_name in self.models.keys():
                if dataset_name is None or dataset_name in model_name:
                    model = self.models[model_name]
                    if best_score[1] > model.state_dict_["loss"][-1].item():
                        best_score[0] = model
                        best_score[1] = model.state_dict_["loss"][-1].item()
                        best_score[2] = model_name
        else:
            raise NotImplementedError(
                f"Invalid scoring, expect infonce_loss, got {scoring}.")

        if best_score[1] == float("inf"):
            raise ValueError(
                f"Invalid dataset name, that dataset wasn't find in the list of fitted models, got {dataset_name}."
            )

        return best_score[0], best_score[2]

    def get_df_results(self, models_dir: str = None) -> pd.DataFrame:
        """Create a :py:class:`pandas.DataFrame` containing the parameters and results for each model.

        Args:
            models_dir: The path to the folder in which save the (fitted) models.

        Returns:
            A :py:class:`pandas.DataFrame` in which each row corresponds to a model from the grid search
            and contains the parameters used for the model, the dataset name that was used for fitting and
            the performances obtained with that model.

        Example:

            >>> import cebra.grid_search
            >>> import numpy as np
            >>> neural_data =  np.random.uniform(0, 1, (300, 30))
            >>> # 1. Define the parameters for the models
            >>> params_grid = dict(
            ...     output_dimension = [3, 16],
            ...     learning_rate = [0.001],
            ...     time_offsets = 5,
            ...     max_iterations = 10,
            ...     verbose = False)
            >>> # 2. Fit the models generated from the list of parameters
            >>> grid_search = cebra.grid_search.GridSearch()
            >>> grid_search = grid_search.fit_models(datasets={"neural_data": neural_data},
            ...                        params=params_grid,
            ...                        models_dir="grid_search_models")
            >>> # 3. Get results for all models
            >>> df_results = grid_search.get_df_results()

        """
        if not hasattr(self, "models_dir"):
            if models_dir is None:
                raise ValueError(
                    "Missing models directory, provide a value for the models_dir parameter."
                )
            self.models_dir = models_dir

        if not hasattr(self, "models") or not hasattr(self, "parameter_grid"):
            self.models, self.parameter_grid = GridSearch().load(
                self.models_dir)

        results = pd.DataFrame(self.parameter_grid)
        results["loss"] = [
            model.state_dict_["loss"][-1].item()
            for model in list(self.models.values())
        ]
        results["dataset_name"] = [
            model_name.split("_")[-1] for model_name in self.models.keys()
        ]

        return results

    def plot_loss_comparison(self,
                             models_dir: str = None,
                             **kwargs) -> matplotlib.axes.Axes:
        """Display the losses for all fitted models present in ``models_dir``.

        Note:
            The method is a wrapper around :py:func:`~.compare_models`, meaning you can provide all
            parameters that are provided to :py:func:`~.compare_models` to that function.

        Args:
            models_dir: The path to the folder in which save the (fitted) models.

        Returns:
            A ``matplotlib.axes.Axes`` on which to generate the plot.

        Example:

            >>> import cebra.grid_search
            >>> import numpy as np
            >>> neural_data =  np.random.uniform(0, 1, (300, 30))
            >>> # 1. Define the parameters for the models
            >>> params_grid = dict(
            ...     output_dimension = [3, 16],
            ...     learning_rate = [0.001],
            ...     time_offsets = 5,
            ...     max_iterations = 10,
            ...     verbose = False)
            >>> # 2. Fit the models generated from the list of parameters
            >>> grid_search = cebra.grid_search.GridSearch()
            >>> grid_search = grid_search.fit_models(datasets={"neural_data": neural_data},
            ...                        params=params_grid,
            ...                        models_dir="grid_search_models")
            >>> # 3. Plot losses for all models
            >>> ax = grid_search.plot_loss_comparison()

        """
        if not hasattr(self, "models_dir"):
            if models_dir is None:
                raise ValueError(
                    "Missing models directory, provide a value for the models_dir parameter."
                )
            self.models_dir = models_dir

        if not hasattr(self, "models"):
            self.models, _ = GridSearch().load(self.models_dir)

        ax = cebra_matplotlib.compare_models(models=list(self.models.values()),
                                             labels=list(self.models.keys()),
                                             **kwargs)
        return ax

    def plot_transform(self):
        """TODO."""
        raise NotImplementedError()



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
import torch
from torch import nn

import cebra.models.criterions as cebra_criterions


@torch.jit.script
def ref_dot_similarity(ref: torch.Tensor, pos: torch.Tensor, neg: torch.Tensor,
                       temperature: float):
    pos_dist = torch.einsum("ni,ni->n", ref, pos) / temperature
    neg_dist = torch.einsum("ni,mi->nm", ref, neg) / temperature
    return pos_dist, neg_dist


@torch.jit.script
def ref_euclidean_similarity(ref: torch.Tensor, pos: torch.Tensor,
                             neg: torch.Tensor, temperature: float):
    ref_sq = torch.einsum("ni->n", ref**2) / temperature
    pos_sq = torch.einsum("ni->n", pos**2) / temperature
    neg_sq = torch.einsum("ni->n", neg**2) / temperature

    pos_cosine, neg_cosine = ref_dot_similarity(ref, pos, neg, temperature)
    pos_dist = -(ref_sq + pos_sq - 2 * pos_cosine)
    neg_dist = -(ref_sq[:, None] + neg_sq[None] - 2 * neg_cosine)

    return pos_dist, neg_dist


@torch.jit.script
def ref_infonce(pos_dist: torch.Tensor, neg_dist: torch.Tensor):
    with torch.no_grad():
        c, _ = neg_dist.max(dim=1, keepdim=True)
    c = c.detach()
    pos_dist = pos_dist - c.squeeze(1)
    neg_dist = neg_dist - c

    align = (-pos_dist).mean()
    uniform = torch.logsumexp(neg_dist, dim=1).mean()
    return align + uniform, align, uniform


@torch.jit.script
def ref_infonce_not_stable(pos_dist: torch.Tensor, neg_dist: torch.Tensor):
    pos_dist = pos_dist
    neg_dist = neg_dist

    align = (-pos_dist).mean()
    uniform = torch.logsumexp(neg_dist, dim=1).mean()
    return align + uniform, align, uniform


class ReferenceInfoNCE(nn.Module):
    """The InfoNCE loss.
    Attributes:
        temperature (float): The softmax temperature
    """

    def __init__(self, temperature: float = 1.0):
        super().__init__()
        self.temperature = temperature

    def _distance(self, ref, pos, neg):
        return ref_dot_similarity(ref, pos, neg, self.temperature)

    def forward(self, ref, pos, neg):
        pos_dist, neg_dist = self._distance(ref, pos, neg)
        return ref_infonce(pos_dist, neg_dist)


class ReferenceInfoMSE(ReferenceInfoNCE):
    """A variant of the InfoNCE loss using a MSE error.
    Attributes:
        temperature (float): The softmax temperature
    """

    def _distance(self, ref, pos, neg):
        return ref_euclidean_similarity(ref, pos, neg, self.temperature)


def setup():
    ref = torch.randn(100).float().unsqueeze(1)
    pos = torch.randn(100).float().unsqueeze(1)
    neg = torch.randn(100).float().unsqueeze(1)
    return ref, pos, neg


@pytest.mark.parametrize("temperature", [0.1, 1.0, 5.0])
@pytest.mark.parametrize(
    "criterion",
    [
        ReferenceInfoNCE,
        ReferenceInfoMSE,
        cebra_criterions.InfoNCE,
        cebra_criterions.InfoMSE,
        cebra_criterions.FixedCosineInfoNCE,
        cebra_criterions.FixedEuclideanInfoNCE,
        cebra_criterions.LearnableCosineInfoNCE,
        cebra_criterions.LearnableEuclideanInfoNCE,
    ],
)
def test_infonce(temperature, criterion):
    """Test infonce loss is computed correctly."""
    ref, pos, neg = setup()
    infonce = criterion(temperature=temperature)
    # perfect alignment
    loss1, _, _ = infonce(ref, ref, neg)
    # random alignment
    loss2, _, _ = infonce(ref, pos, neg)
    assert loss1 < loss2

    infonce_lower_temp = criterion(temperature=temperature / 2.0)
    infonce_higher_temp = criterion(temperature=temperature * 2.0)
    loss, _, _ = infonce(ref, pos, neg)
    loss_low, _, _ = infonce_lower_temp(ref, pos, neg)
    loss_high, _, _ = infonce_higher_temp(ref, pos, neg)
    assert not torch.allclose(loss, loss_high)
    assert not torch.allclose(loss, loss_low)


@pytest.mark.parametrize("temperature", [0.1, 1.0, 5.0])
@pytest.mark.parametrize(
    "criterion",
    [
        ReferenceInfoNCE,
        cebra_criterions.FixedCosineInfoNCE,
        cebra_criterions.FixedEuclideanInfoNCE,
    ],
)
def test_fixed(temperature, criterion):
    # Check that temperature is frozen when trainable option is False
    ref, pos, neg = setup()
    infonce_fixed = criterion(temperature=temperature)
    assert len(list(infonce_fixed.parameters())) == 0


@pytest.mark.parametrize("temperature", [0.1, 1.0, 5.0])
@pytest.mark.parametrize(
    "criterion",
    [
        cebra_criterions.LearnableCosineInfoNCE,
        cebra_criterions.LearnableEuclideanInfoNCE,
    ],
)
def test_trainable(temperature, criterion):
    # Check that temperature is trainable
    ref, pos, neg = setup()
    infonce_learnable = criterion(temperature=temperature)
    (temp,) = list(iter(infonce_learnable.parameters()))
    loss, _, _ = infonce_learnable(ref, pos, neg)
    loss.backward()
    assert temp.grad is not None


def test_clipping():
    # Check that clipping works
    ref, pos, neg = setup()
    clipped_infonce = cebra_criterions.LearnableCosineInfoNCE(
        temperature=100, min_temperature=0.1)
    nonclipped_infonce = cebra_criterions.LearnableCosineInfoNCE(
        temperature=100, min_temperature=None)
    assert torch.allclose(
        clipped_infonce(ref, pos, neg)[0],
        nonclipped_infonce(ref, pos, neg)[0])


@pytest.mark.parametrize("temperature", [0.1, 1.0, 5.0])
def test_infonce_equivalence(temperature):
    # Check if InfoNCE with non trainable option is equivalent to old implementation
    ref, pos, neg = setup()
    infonce_old = cebra_criterions.InfoNCE(temperature=temperature)
    infonce_learnable = cebra_criterions.LearnableCosineInfoNCE(
        temperature=temperature)
    loss_old, _, _ = infonce_old(ref, pos, neg)
    loss_learnable, _, _ = infonce_learnable(ref, pos, neg)
    assert torch.allclose(loss_old,
                          loss_learnable.detach(),
                          rtol=1e-04,
                          atol=1e-05)
    assert len(list(infonce_old.parameters())) == 0
    assert len(list(infonce_learnable.parameters())) == 1


@pytest.mark.parametrize("temperature", [0.1, 1.0, 5.0])
def test_infonce_reference_new_equivalence(temperature):
    # Check if InfoNCE with non trainable option is equivalent to reference implementation
    ref, pos, neg = setup()
    ref_infonce = ReferenceInfoNCE(temperature=temperature)
    new_infonce = cebra_criterions.InfoNCE(temperature=temperature)
    ref_infomse = ReferenceInfoMSE(temperature=temperature)
    new_infomse = cebra_criterions.InfoMSE(temperature=temperature)
    cosine_loss_old, _, _ = ref_infonce(ref, pos, neg)
    euclidean_loss_old, _, _ = ref_infomse(ref, pos, neg)
    cosine_loss_new, _, _ = new_infonce(ref, pos, neg)
    euclidean_loss_new, _, _ = new_infomse(ref, pos, neg)
    assert torch.allclose(cosine_loss_old,
                          cosine_loss_new,
                          rtol=1e-04,
                          atol=1e-05)
    assert torch.allclose(euclidean_loss_old,
                          euclidean_loss_new,
                          rtol=1e-04,
                          atol=1e-05)


def test_alias():
    assert cebra_criterions.InfoNCE == cebra_criterions.FixedCosineInfoNCE
    assert cebra_criterions.InfoMSE == cebra_criterions.FixedEuclideanInfoNCE


def _reference_dot_similarity(ref, pos, neg):
    pos_dist = torch.zeros(ref.shape[0])
    neg_dist = torch.zeros(ref.shape[0], neg.shape[0])
    for d in range(ref.shape[1]):
        for i in range(len(ref)):
            pos_dist[i] += ref[i, d] * pos[i, d]
            for j in range(len(neg)):
                neg_dist[i, j] += ref[i, d] * neg[j, d]
    return pos_dist, neg_dist


def _reference_euclidean_similarity(ref, pos, neg):
    pos_dist = torch.zeros(ref.shape[0])
    neg_dist = torch.zeros(ref.shape[0], neg.shape[0])
    for d in range(ref.shape[1]):
        for i in range(len(ref)):
            pos_dist[i] += -(ref[i, d] - pos[i, d])**2
            for j in range(len(neg)):
                neg_dist[i, j] += -(ref[i, d] - neg[j, d])**2
    return pos_dist, neg_dist


def _reference_infonce(pos_dist, neg_dist):
    align = -pos_dist.mean()
    uniform = torch.logsumexp(neg_dist, dim=1).mean()
    return align + uniform, align, uniform


def test_similiarities():
    rng = torch.Generator().manual_seed(42)
    ref = torch.randn(10, 3, generator=rng)
    pos = torch.randn(10, 3, generator=rng)
    neg = torch.randn(12, 3, generator=rng)

    pos_dist, neg_dist = _reference_dot_similarity(ref, pos, neg)
    pos_dist_2, neg_dist_2 = cebra_criterions.dot_similarity(ref, pos, neg)

    assert torch.allclose(pos_dist, pos_dist_2)
    assert torch.allclose(neg_dist, neg_dist_2)

    pos_dist, neg_dist = _reference_euclidean_similarity(ref, pos, neg)
    pos_dist_2, neg_dist_2 = cebra_criterions.euclidean_similarity(
        ref, pos, neg)

    assert torch.allclose(pos_dist, pos_dist_2)
    assert torch.allclose(neg_dist, neg_dist_2)


def _compute_grads(output, inputs):
    for input_ in inputs:
        input_.grad = None
        assert input_.requires_grad
    output.backward()
    return [input_.grad for input_ in inputs]


def _sample_dist_matrices(seed):
    rng = torch.Generator().manual_seed(42)
    pos_dist = torch.randn(100, generator=rng)
    neg_dist = torch.randn(100, 100, generator=rng)
    return pos_dist, neg_dist


@pytest.mark.parametrize("seed", [42, 4242, 424242])
def test_infonce_check_output_parts(seed):
    pos_dist, neg_dist = _sample_dist_matrices(seed)

    ref_loss, ref_align, ref_uniform = _reference_infonce(pos_dist, neg_dist)
    loss, align, uniform = cebra_criterions.infonce(pos_dist, neg_dist)

    assert torch.allclose(ref_loss, loss)
    assert torch.allclose(ref_align, align, atol=0.0001)
    assert torch.allclose(ref_uniform, uniform)
    assert torch.allclose(align + uniform, loss)


@pytest.mark.parametrize("seed", [42, 4242, 424242])
@pytest.mark.parametrize("case", [0, 1, 2])
def test_infonce_gradients(seed, case):
    pos_dist, neg_dist = _sample_dist_matrices(seed)

    # TODO(stes): This test seems to fail due to some recent software
    # updates; root cause not identified. Remove this comment once
    # fixed. (for i = 0, 1)
    pos_dist_ = pos_dist.clone()
    neg_dist_ = neg_dist.clone()
    pos_dist_.requires_grad_(True)
    neg_dist_.requires_grad_(True)
    loss_ref = _reference_infonce(pos_dist_, neg_dist_)[case]
    grad_ref = _compute_grads(loss_ref, [pos_dist_, neg_dist_])

    pos_dist_ = pos_dist.clone()
    neg_dist_ = neg_dist.clone()
    pos_dist_.requires_grad_(True)
    neg_dist_.requires_grad_(True)
    loss = cebra_criterions.infonce(pos_dist_, neg_dist_)[case]
    grad = _compute_grads(loss, [pos_dist_, neg_dist_])

    # NOTE(stes) default relative tolerance is 1e-5
    assert torch.allclose(loss_ref, loss, rtol=1e-4)

    if case == 0:
        assert grad[0] is not None
        assert grad[1] is not None
        assert torch.allclose(grad_ref[0], grad[0])
        assert torch.allclose(grad_ref[1], grad[1])
    if case == 1:
        assert grad[0] is not None
        assert torch.allclose(grad_ref[0], grad[0])
        # TODO(stes): This is most likely not the right fix, needs more
        # investigation. On the first run of the test, grad[1] is actually
        # None, and then on the second run of the test it is a Tensor, but
        # with zeros everywhere. The behavior is fine for fitting models,
        # but there is some side-effect in our test suite we need to fix.
        if grad[1] is not None:
            assert torch.allclose(grad[1], torch.zeros_like(grad[1]))
    if case == 2:
        if grad[0] is not None:
            assert torch.allclose(grad[0], torch.zeros_like(grad[0]))
        assert grad[1] is not None
        assert torch.allclose(grad_ref[1], grad[1])



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This package contains abstract base classes for different solvers.

Solvers are used to package models, criterions and optimizers and implement training
loops. When subclassing abstract solvers, in the simplest case only the
:py:meth:`Solver._inference` needs to be overridden.

For more complex use cases, the :py:meth:`Solver.step` and
:py:meth:`Solver.fit` method can be overridden to
implement larger changes to the training loop.
"""

import abc
import os
from typing import Callable, Dict, List, Literal, Optional

import literate_dataclasses as dataclasses
import torch

import cebra
import cebra.data
import cebra.io
import cebra.models
from cebra.solver.util import Meter
from cebra.solver.util import ProgressBar


@dataclasses.dataclass
class Solver(abc.ABC, cebra.io.HasDevice):
    """Solver base class.

    A solver contains helper methods for bundling a model, criterion and optimizer.

    Attributes:
        model: The encoder for transforming reference, positive and negative samples.
        criterion: The criterion computed from the similarities between positive pairs
            and negative pairs. The criterion can have trainable parameters on its own.
        optimizer: A PyTorch optimizer for updating model and criterion parameters.
        history: Deprecated since 0.0.2. Use :py:attr:`log`.
        decode_history: Deprecated since 0.0.2. Use a hook during training for validation and
            decoding. See the arguments of :py:meth:`fit`.
        log: The logs recorded during training, typically contains the ``total`` loss as well
            as the logs for positive (``pos``) and negative (``neg``) pairs. For the standard
            criterions in CEBRA, also contains the value of the ``temperature``.
        tqdm_on: Use ``tqdm`` for showing a progress bar during training.
    """

    model: torch.nn.Module
    criterion: torch.nn.Module
    optimizer: torch.optim.Optimizer
    history: List = dataclasses.field(default_factory=list)
    decode_history: List = dataclasses.field(default_factory=list)
    log: Dict = dataclasses.field(default_factory=lambda: ({
        "pos": [],
        "neg": [],
        "total": [],
        "temperature": []
    }))
    tqdm_on: bool = True

    def __post_init__(self):
        cebra.io.HasDevice.__init__(self)
        self.best_loss = float("inf")

    def state_dict(self) -> dict:
        """Return a dictionary fully describing the current solver state.

        Returns:
            State dictionary, including the state dictionary of the models and
            optimizer. Also contains the training history and the CEBRA version
            the model was trained with.
        """

        return {
            "model": self.model.state_dict(),
            "optimizer": self.optimizer.state_dict(),
            "loss": torch.tensor(self.history),
            "decode": self.decode_history,
            "criterion": self.criterion.state_dict(),
            "version": cebra.__version__,
            "log": self.log,
        }

    def load_state_dict(self, state_dict: dict, strict: bool = True):
        """Update the solver state with the given state_dict.

        Args:
            state_dict: Dictionary with parameters for the `model`, `optimizer`,
                and the past loss history for the solver.
            strict: Make sure all states can be loaded. Set to `False` to allow
                to partially load the state for all given keys.
        """

        def _contains(key):
            if key in state_dict:
                return True
            elif strict:
                raise KeyError(
                    f"Key {key} missing in state_dict. Contains: {list(state_dict.keys())}."
                )
            return False

        def _get(key):
            return state_dict.get(key)

        if _contains("model"):
            self.model.load_state_dict(_get("model"))
        if _contains("criterion"):
            self.criterion.load_state_dict(_get("criterion"))
        if _contains("optimizer"):
            self.optimizer.load_state_dict(_get("optimizer"))
        # TODO(stes): This will be deprecated at some point; the "log" attribute
        # holds the same information.
        if _contains("loss"):
            self.history = _get("loss").cpu().numpy().tolist()
        if _contains("decode"):
            self.decode_history = _get("decode")
        if _contains("log"):
            self.log = _get("log")

    @property
    def num_parameters(self) -> int:
        """Total number of parameters in the encoder and criterion."""
        return sum(p.numel() for p in self.parameters())

    def parameters(self):
        """Iterate over all parameters."""
        for parameter in self.model.parameters():
            yield parameter

        for parameter in self.criterion.parameters():
            yield parameter

    def _get_loader(self, loader):
        return ProgressBar(
            loader,
            "tqdm" if self.tqdm_on else "off",
        )

    def fit(
        self,
        loader: cebra.data.Loader,
        valid_loader: cebra.data.Loader = None,
        *,
        save_frequency: int = None,
        valid_frequency: int = None,
        decode: bool = False,
        logdir: str = None,
        save_hook: Callable[[int, "Solver"], None] = None,
    ):
        """Train model for the specified number of steps.

        Args:
            loader: Data loader, which is an iterator over `cebra.data.Batch` instances.
                Each batch contains reference, positive and negative input samples.
            valid_loader: Data loader used for validation of the model.
            save_frequency: If not `None`, the frequency for automatically saving model checkpoints
                to `logdir`.
            valid_frequency: The frequency for running validation on the ``valid_loader`` instance.
            logdir:  The logging directory for writing model checkpoints. The checkpoints
                can be read again using the `solver.load` function, or manually via loading the
                state dict.

        TODO:
            * Refine the API here. Drop the validation entirely, and implement this via a hook?
        """

        self.to(loader.device)

        iterator = self._get_loader(loader)
        self.model.train()
        for num_steps, batch in iterator:
            stats = self.step(batch)
            iterator.set_description(stats)

            if save_frequency is None:
                continue
            save_model = num_steps % save_frequency == 0
            run_validation = (valid_loader
                              is not None) and (num_steps % valid_frequency
                                                == 0)
            if run_validation:
                validation_loss = self.validation(valid_loader)
                if self.best_loss is None or validation_loss < self.best_loss:
                    self.best_loss = validation_loss
                    self.save(logdir, "checkpoint_best.pth")
            if save_model:
                if decode:
                    self.decode_history.append(
                        self.decoding(loader, valid_loader))
                if save_hook is not None:
                    save_hook(num_steps, self)
                self.save(logdir, f"checkpoint_{num_steps:#07d}.pth")

    def step(self, batch: cebra.data.Batch) -> dict:
        """Perform a single gradient update.

        Args:
            batch: The input samples

        Returns:
            Dictionary containing training metrics.
        """
        self.optimizer.zero_grad()
        prediction = self._inference(batch)
        loss, align, uniform = self.criterion(prediction.reference,
                                              prediction.positive,
                                              prediction.negative)

        loss.backward()
        self.optimizer.step()
        self.history.append(loss.item())
        stats = dict(
            pos=align.item(),
            neg=uniform.item(),
            total=loss.item(),
            temperature=self.criterion.temperature,
        )
        for key, value in stats.items():
            self.log[key].append(value)
        return stats

    def validation(self,
                   loader: cebra.data.Loader,
                   session_id: Optional[int] = None):
        """Compute score of the model on data.

        Args:
            loader: Data loader, which is an iterator over `cebra.data.Batch` instances.
                Each batch contains reference, positive and negative input samples.
            session_id: The session ID, an integer between 0 and the number of sessions in the
                multisession model, set to None for single session.

        Returns:
            Loss averaged over iterations on data batch.
        """
        assert (session_id is None) or (session_id == 0)
        iterator = self._get_loader(loader)
        total_loss = Meter()
        self.model.eval()
        for _, batch in iterator:
            prediction = self._inference(batch)
            loss, _, _ = self.criterion(prediction.reference,
                                        prediction.positive,
                                        prediction.negative)
            total_loss.add(loss.item())
        return total_loss.average

    @torch.no_grad()
    def decoding(self, train_loader, valid_loader):
        """Deprecated since 0.0.2."""
        train_x = self.transform(train_loader.dataset[torch.arange(
            len(train_loader.dataset.neural))])
        train_y = train_loader.dataset.index
        valid_x = self.transform(valid_loader.dataset[torch.arange(
            len(valid_loader.dataset.neural))])
        valid_y = valid_loader.dataset.index
        decode_metric = train_loader.dataset.decode(
            train_x.cpu().numpy(),
            train_y.cpu().numpy(),
            valid_x.cpu().numpy(),
            valid_y.cpu().numpy(),
        )
        return decode_metric

    @torch.no_grad()
    def transform(self, inputs: torch.Tensor) -> torch.Tensor:
        """Compute the embedding.

        This function by default only applies the ``forward`` function
        of the given model, after switching it into eval mode.

        Args:
            inputs: The input signal

        Returns:
            The output embedding.

        TODO:
            * Remove eval mode
        """

        self.model.eval()
        return self.model(inputs)

    @abc.abstractmethod
    def _inference(self, batch: cebra.data.Batch) -> cebra.data.Batch:
        """Given a batch of input examples, return the model outputs.

        TODO: make this a public function?

        Args:
            batch: The input data, not necessarily aligned across the batch
                dimension. This means that ``batch.index`` specifies the map
                between reference/positive samples, if not equal ``None``.

        Returns:
            Processed batch of data. While the input data might not be aligned
            across the sample dimensions, the output data should be aligned and
            ``batch.index`` should be set to ``None``.
        """
        raise NotImplementedError

    def load(self, logdir, filename="checkpoint.pth"):
        """Load the experiment from its checkpoint file.

        Args:
            filename (str): Checkpoint name for loading the experiment.
        """

        savepath = os.path.join(logdir, filename)
        if not os.path.exists(savepath):
            print("Did not find a previous experiment. Starting from scratch.")
            return
        checkpoint = torch.load(savepath, map_location=self.device)
        self.load_state_dict(checkpoint, strict=True)

    def save(self, logdir, filename="checkpoint_last.pth"):
        """Save the model and optimizer params.

        Args:
            logdir: Logging directory for this model.
            filename: Checkpoint name for saving the experiment.
        """
        if not os.path.exists(os.path.dirname(logdir)):
            os.makedirs(logdir)
        savepath = os.path.join(logdir, filename)
        torch.save(
            self.state_dict(),
            savepath,
        )


@dataclasses.dataclass
class MultiobjectiveSolver(Solver):
    """Train models to satisfy multiple learning objectives.

    This variant of the standard :py:class:`cebra.solver.base.Solver` implements multi-objective
    or "hybrid" training.

    Attributes:
        model: A multi-objective CEBRA model
        optimizer: The optimizer used for training.
        num_behavior_features: The feature dimension for the features dedicated
            to satisfy the behavior contrastive objective. The remainder is used
            for time contrastive learning.
        renormalize_features: If ``True``, normalize the behavior and time
            contrastive features individually before computing similarity scores.
    """

    num_behavior_features: int = 3
    renormalize_features: bool = False
    output_mode: Literal["overlapping", "separate"] = "overlapping"

    @property
    def num_time_features(self):
        return self.num_total_features - self.num_behavior_features

    @property
    def num_total_features(self):
        return self.model.num_output

    def __post_init__(self):
        super().__post_init__()
        self._check_dimensions()
        self.model = cebra.models.MultiobjectiveModel(
            self.model,
            dimensions=(self.num_behavior_features, self.model.num_output),
            renormalize=self.renormalize_features,
            output_mode=self.output_mode,
        )

    def _check_dimensions(self):
        """Check the feature dimensions for behavior/time contrastive learning.

        Raises:
            ValueError: If feature dimensions are larger than the model features,
                or not sufficiently large for renormalization.
        """
        if self.output_mode == "separate":
            if self.num_behavior_features >= self.num_total_features:
                raise ValueError(
                    "For multi-objective training, the number of features for "
                    f"behavior contrastive learning ({self.num_behavior_features}) cannot be as large or larger "
                    f"than the total feature dimension ({self.num_total_features})."
                )
            if self.num_time_features >= self.num_total_features:
                raise ValueError(
                    "For multi-objective training, the number of features for "
                    f"time contrastive learning ({self.num_time_features}) cannot be as large or larger "
                    f"than the total feature dimension ({self.num_total_features})."
                )
        if self.renormalize_features:
            if self.num_behavior_features < 2:
                raise ValueError(
                    "When renormalizing the features, the feature dimension needs "
                    "to be at least 2 for behavior. "
                    "Check the values of 'renormalize_features' and 'num_behavior_features'."
                )
            if self.num_time_features < 2:
                raise ValueError(
                    "When renormalizing the features, the feature dimension needs "
                    "to be at least 2 for behavior. "
                    "Check the values of 'renormalize_features' and 'num_time_features'."
                )

    def step(self, batch: cebra.data.Batch) -> dict:
        """Perform a single gradient update with multiple objectives.

        Args:
            batch: The input samples

        Returns:
            Dictionary containing training metrics.
        """
        self.optimizer.zero_grad()
        prediction_behavior, prediction_time = self._inference(batch)

        behavior_loss, behavior_align, behavior_uniform = self.criterion(
            prediction_behavior.reference,
            prediction_behavior.positive,
            prediction_behavior.negative,
        )

        time_loss, time_align, time_uniform = self.criterion(
            prediction_time.reference,
            prediction_time.positive,
            prediction_time.negative,
        )

        loss = behavior_loss + time_loss
        loss.backward()
        self.optimizer.step()
        self.history.append(loss.item())
        return dict(
            behavior_pos=behavior_align.item(),
            behavior_neg=behavior_uniform.item(),
            behavior_total=behavior_loss.item(),
            time_pos=time_align.item(),
            time_neg=time_uniform.item(),
            time_total=time_loss.item(),
        )



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Distributions for sampling from continuously indexed datasets."""

from typing import Literal, Optional

import torch

import cebra.data
import cebra.distributions
import cebra.distributions.base as abc_
from cebra.data.datatypes import Offset


class Prior(abc_.PriorDistribution, abc_.HasGenerator):
    """An empirical prior distribution for continuous datasets.

    Given the index, uniformly sample across time steps, i.e.,
    sample from the empirical distribution.

    Args:
        continuous: The multi-dimensional continuous index.
    """

    def __init__(self,
                 continuous: torch.Tensor,
                 device: Literal["cpu", "cuda"] = "cpu",
                 seed: int = 0):
        abc_.HasGenerator.__init__(self, device=device, seed=seed)
        self.continuous = continuous
        self.num_samples = len(self.continuous)

    def sample_prior(self,
                     num_samples: int,
                     offset: Optional[Offset] = None) -> torch.Tensor:
        """Return uniformly sampled indices.

        Args:
            num_samples: The number of samples to draw from the prior
                distribution. This will be the length of the returned
                tensor.
            offset: The :py:class:`cebra.data.datatypes.Offset` offset
                to be respected when sampling indices. The minimum index
                sampled will be ``offset.left`` (inclusive), the maximum
                index will be the index length minus ``offset.right``
                (exclusive).

        Returns:
            An integer tensor of shape ``num_samples`` containing
            random indices
        """

        if offset is not None and self.num_samples <= len(self.offset):
            raise ValueError(
                f"Dataset is too small, got {self.num_samples} timepoints "
                f"to train a model with offset {self.offset}.")

        return self.randint(
            0 if offset is None else self.offset.left,
            self.num_samples - (0 if offset is None else self.offset.right),
            (num_samples,),
        )


class TimeContrastive(abc_.JointDistribution, abc_.HasGenerator):
    """Time contrastive learning.

    Positive samples will have a distance of exactly :py:attr:`time_offset`
    samples in time.

    Attributes:
        continuous: The multi-dimensional continuous index.
        time_offset: The time delay between samples that form a positive pair
        num_samples: TODO(stes) remove?
        device: Device (cpu or gpu)
        seed: The seed for sampling from the prior and negative distribution
            (TODO currentlty not used)

    TODO:
        Implement time contrastive learning across multiple time steps, e.g.
        by sampling the time offset in the conditional distribution.
    """

    def __init__(
        self,
        continuous: Optional[torch.Tensor] = None,
        time_offset: int = 1,
        num_samples: Optional[int] = None,
        device: Literal["cpu", "cuda"] = "cpu",
        seed: Optional[int] = None,
    ):
        abc_.HasGenerator.__init__(self, device=device, seed=seed)
        if continuous is None and num_samples is None:
            raise ValueError(
                "Supply either a continuous index (which will be used to infer the dataset size) "
                "or alternatively the number of datapoints using the num_samples argument."
            )
        if continuous is not None and num_samples is not None:
            if len(continuous) != num_samples:
                raise ValueError(
                    f"It is sufficient to provide either the continuous index, or the num_samples "
                    f"argument. You provided both, and the length of continuous ({len(continuous)} does "
                    f"not match num_samples={num_samples} you provided.")
        self.num_samples = len(
            continuous) if num_samples is None else num_samples
        self.time_offset = time_offset
        if self.num_samples <= self.time_offset:
            raise ValueError(
                f"number of samples has to exceed the time offset, but got {self.num_samples} <= {self.time_offset}."
            )

    def sample_prior(self,
                     num_samples: int,
                     offset: Optional[Offset] = None) -> torch.Tensor:
        """Return a random index sample, respecting the given time offset.

        Prior samples are uniformly sampled from ``[0, T - t)`` where ``T`` is the total
        number of samples in the index, and ``t`` is the time offset used for sampling.

        Args:
            num_samples: Number of time steps to draw uniformly from the
                number of available time steps in the dataset
            offset: The model offset to respect for sampling from the prior.
                TODO not yet implemented

        Returns:
            A ``(num_samples,)`` shaped tensor containing time indices from the
            uniform prior distribution.
        """
        if offset is not None:
            raise NotImplementedError(
                f"It is not yet supported to use an offset for sampling from the index "
                f"prior for time contrastive learning. The offset is fixed to "
                f"the specified time_offset={self.time_offset}")
        return self.randint(0, self.num_samples - self.time_offset,
                            (num_samples,))

    def sample_conditional(self, reference_idx: torch.Tensor) -> torch.Tensor:
        """Return samples from the time-contrastive conditional distribution.

        The returned indices will be given by incrementing the reference indices
        by the specified :py:attr:`time_offset`. When the reference indices are
        sampled with :py:meth:`sample_prior`, it is ensured that the indices all
        lie within the bounds of the dataset.

        Args:
            reference_idx: The time indices of the reference samples

        Returns:
            A ``(len(reference_idx),)`` shaped tensor containing time indices from the
            time-contrastive conditional distribution. The samples will be simply
            offset by :py:attr:`time_offset` from ``reference_idx``.
        """
        return reference_idx + self.time_offset


class DirectTimedeltaDistribution(TimeContrastive, abc_.HasGenerator):
    """Look up indices with

    Todo:
        - This class is work in progress.
    """

    def __init__(self, continuous: torch.Tensor, time_offset: int = 1):
        super().__init__(continuous=continuous, time_offset=time_offset)
        self.index = cebra.distributions.ContinuousIndex(self.data)

    def sample_conditional(self, reference_idx: torch.Tensor) -> torch.Tensor:
        """Samples from the conditional distribution.

        Todo:
            - class and this function is work in progress.
        """
        query_idx = super().sample_conditional(reference_idx)
        query = self.index[query_idx]
        # TODO(stes): This will by default simply return the query_idx. This should
        # be covered by a test and fixed for a future release.
        return self.index.search(query)


class TimedeltaDistribution(abc_.JointDistribution, abc_.HasGenerator):
    """Define a conditional distribution based on behavioral changes over time.

    Takes a continuous index, and uses the empirical distribution of differences
    between samples in this index.

    Args:
        continuous: The multidimensional, continuous index
        time_delta: The time delay between samples that should form a positive
            pair.
        device: TODO
        seed: TODO

    Note:
        For best results, the given continuous index should contain independent
        factors; positive pairs will be formed by adding a _random_ difference
        estimated within the dataset to the reference samples.
        Factors should ideally also be within the same range (since the Euclidean
        distance is used in the search). A simple solution is to perform a PCA
        or ICA, or apply CEBRA first before passing the index to this function.
    """

    def __init__(self,
                 continuous,
                 time_delta: int = 1,
                 device: Literal["cpu", "cuda"] = "cpu",
                 seed: Optional[int] = None):
        abc_.HasGenerator.__init__(self, device=device, seed=seed)
        self.data = continuous
        self.time_delta = time_delta
        self.time_difference = torch.zeros_like(self.data, device=self.device)
        self.time_difference[time_delta:] = (self.data[time_delta:] -
                                             self.data[:-time_delta])
        self.index = cebra.distributions.ContinuousIndex(self.data)
        self.prior = Prior(self.data, device=device, seed=seed)

    def sample_prior(self, num_samples: int) -> torch.Tensor:
        """See :py:meth:`.Prior.sample_prior`."""
        return self.prior.sample_prior(num_samples)

    def sample_conditional(self, reference_idx: torch.Tensor) -> torch.Tensor:
        """Return indices from the conditional distribution."""

        if reference_idx.dim() != 1:
            raise ValueError(
                f"Reference indices have wrong shape: {reference_idx.shape}. "
                "Pass a 1D array of indices of reference samples.")

        num_samples = reference_idx.size(0)
        diff_idx = self.randint(len(self.time_difference), (num_samples,))
        query = self.data[reference_idx] + self.time_difference[diff_idx]
        return self.index.search(query)


class DeltaNormalDistribution(abc_.JointDistribution, abc_.HasGenerator):
    """Define a conditional distribution based on behavioral changes over time.

    Takes a continuous index, and uses sample from Gaussian distribution to sample positive pairs.
    Note that if the continuous index is multidimensional, the Gaussian distribution will have
    isotropic covariance matrix i.e. Σ = sigma^2 * I.

    Args:
        continuous: The multidimensional, continuous index.
        delta: Standard deviation of Gaussian distribution to sample positive pair.

    """

    def __init__(self,
                 continuous: torch.Tensor,
                 delta: float = 0.1,
                 device: Literal["cpu", "cuda"] = "cpu",
                 seed: Optional[int] = None):
        abc_.HasGenerator.__init__(self, device=device, seed=seed)
        self.data = continuous
        self.std = delta
        self.index = cebra.distributions.ContinuousIndex(self.data)
        self.prior = Prior(self.data, device=device, seed=seed)

    def sample_prior(self, num_samples: int) -> torch.Tensor:
        """See :py:meth:`.Prior.sample_prior`."""
        return self.prior.sample_prior(num_samples)

    def sample_conditional(self, reference_idx: torch.Tensor) -> torch.Tensor:
        """Return indices from the conditional distribution."""

        if reference_idx.dim() != 1:
            raise ValueError(
                f"Reference indices have wrong shape: {reference_idx.shape}. "
                "Pass a 1D array of indices of reference samples.")

        # TODO(stes): Set seed
        mean = self.data[reference_idx]
        query = torch.distributions.Normal(
            loc=mean,
            scale=torch.ones_like(mean, device=self.device) * self.std,
        ).sample()

        query = query.unsqueeze(-1) if query.dim() == 1 else query
        return self.index.search(query)


class CEBRADistribution(abc_.JointDistribution):
    """Use CEBRA embeddings for defining a conditional distribution.

    TODO:
        - This class is not implemented yet. Contributions welcome!
    """

    pass


def _interleave(tensor: torch.Tensor, N: int) -> torch.Tensor:
    size = tensor.size()
    return tensor.reshape(N, N, -1).transpose(1, 0).reshape(*size)
